"""
Set parameters
"""
from numbers import Number
import numpy as np
import pandas as pd
import sciris as sc
import starsim as ss
import datetime as dt

__all__ = ['Pars', 'SimPars']

# Define classes to not descend into further -- based on sciris.sc_nested
atomic_classes = (str, Number, list, np.ndarray, pd.Series, pd.DataFrame, type(None), dt.date)


class Pars(sc.objdict):
    """
    Dict-like container of parameters

    Acts like an `sc.objdict()`, except that adding new keys are disallowed by
    default, and auto-updates known types.
    """
    def __init__(self, pars=None, **kwargs):
        if pars is not None:
            if isinstance(pars, dict):
                kwargs = pars | kwargs
            else:
                errormsg = f'Cannot supply parameters as type {type(pars)}: must be a dict'
                raise ValueError(errormsg)
        super().__init__(**kwargs)
        return

    def update(self, pars=None, create=False, **kwargs):
        """
        Update internal dict with new pars.

        Args:
            pars (dict): the parameters to update (if None, do nothing)
            create (bool): if create is False, then raise a KeyNotFoundError if the key does not already exist
            kwargs (dict): merged with pars
        """
        # Get parameters, and return if none
        pars = {} if pars is None else dict(pars) # Make it a simple dict, since just iterating over
        pars = pars | kwargs # Merge dictionaries
        if not len(pars): # Check that new parameters are supplied, and if not return unchanged
            return self
        if not create: # Check if there are any mismatches, which raises an exception if there are
            self.check_key_mismatch(pars)

        # Perform the update
        for key,new in pars.items():
            if key not in self.keys(): # It's a new parameter and create=True: update directly
                self[key] = new
            else:
                old = self[key] # Get the existing object we're about to update
                if isinstance(old, atomic_classes): # It's a number, string, etc: update directly
                    self[key] = new

                elif isinstance(old, Pars): # It's a Pars object: update recursively
                    old.update(new, create=create)
                elif isinstance(old, ss.ndict): # Update module containers -- main use case
                    self._update_ndict(key, old, new)
                elif isinstance(old, ss.Module): # Update modules -- rare
                    self._update_module(key, old, new)
                elif isinstance(old, ss.TimePar):
                    self._update_timepar(key, old, new)
                elif isinstance(old, ss.Dist): # Update a distribution
                    self._update_dist(key, old, new)
                elif callable(old): # It's a function: update directly
                    self[key] = new
                elif isinstance(old, dict):
                    self[key] = new # Take dictionaries directly, without warning the user
                elif isinstance(old, (ss.dur, ss.Rate)):
                    self[key] = new
                else: # Everything else; not used currently but could be
                    warnmsg = f'No known mechanism for handling {type(old)} → {type(new)}; using default'
                    ss.warn(warnmsg)
                    self[key] = new
        return self

    def _update_ndict(self, key, old, new):
        """ Update an ndict object in the parameters, e.g. sim.pars.diseases """
        if not len(old): # It's empty, just overwrite
            self[key] = new
        else: # Don't overwrite an existing ndict
            if isinstance(new, dict):
                for newkey,newvals in new.items():
                    old[newkey].pars.update(newvals) # e.g. pars = {'diseases": {'sir': {'dur_inf': 6}}}
            else:
                errormsg = f'Cannot update an ndict with {type(new)}: must be a dict to set new parameters'
                raise TypeError(errormsg)
        return

    def _update_module(self, key, old, new):
        """ Update a Module object in the parameters, e.g. sim.pars.diseases.sir """
        if isinstance(new, dict):
            old[key].pars.update(new) # e.g. pars = {'dur_inf': 6}
        else:
            errormsg = f'Cannot update a module with {type(new)}: must be a dict to set new parameters'
            raise TypeError(errormsg)
        return

    def _update_timepar(self, key, old, new):
        """ Update a time parameter """

        # It's a TimePar, e.g. dur_inf = ss.years(6); use directly
        if isinstance(new, ss.TimePar):
            self[key] = new

        # It's a dataframe, allow the update -- used for demographics
        elif isinstance(new, (pd.Series, pd.DataFrame)):
            self[key] = new # TODO: add validation or convert to timepar

        # It's a single number, e.g. dur_inf = 6; set parameters
        elif sc.isnumber(new):
            old.value = new

        # It's array-like; set the value, but convert first to handle lists etc
        elif sc.checktype(new, 'arraylike'):
            old.value = sc.toarray(new)

        # It's a dict, e.g. beta for different networks, just set it directly and trust the user
        elif isinstance(new, dict):
            # for key,val in new.items(): # TODO: should we include validation like this?
            #     if sc.isnumber(val):
            #         new[key] = old.__class__(val)
            self[key] = new

        # Give up
        else:
            errormsg = f'Updating timepar {old} from {type(old)} to {type(new)} is not supported'
            raise TypeError(errormsg)

        return

    def _update_dist(self, key, old, new):
        """ Update a Dist object in the parameters, e.g. sim.pars.diseases.sir.dur_inf """

        # It's a Dist, e.g. dur_inf = ss.normal(6,2); use directly
        if isinstance(new, ss.Dist):
            if isinstance(old, ss.bernoulli) and not isinstance(new, ss.bernoulli):
                errormsg = f"Bernoulli distributions can't be changed to another type: {type(new)} is invalid"
                raise TypeError(errormsg)
            else:
                self[key] = new

        # It's a single number, e.g. dur_inf = 6; set parameters, or a time parameter
        elif isinstance(new, (Number, ss.dur, ss.Rate)):
            old.set(new)

        # It's a list of numbers, e.g. dur_inf = [6, 2]; set parameters
        elif isinstance(new, list):
            old.set(*new)

        # It's a dict, figure out what to do
        elif isinstance(new, dict):
            newtype = new.get('type')
            if newtype is None: # Same type of dist, set parameters
                old.set(**new)
            else: # We need to create a new distribution
                if isinstance(old, ss.bernoulli) and newtype != 'bernoulli':
                    errormsg = f"Bernoulli distributions can't be changed to another type: {newtype} is invalid"
                    raise TypeError(errormsg)
                else:
                    dist = ss.make_dist(new)
                    self[key] = dist

        # It's a function, treat it like a number
        elif sc.isfunc(new):
            old.set(new)

        # Give up
        else:
            errormsg = f'Updating dist from {type(old)} to {type(new)} is not supported'
            raise TypeError(errormsg)

        return

    def check_key_mismatch(self, pars):
        """ Check whether additional keys are being added to the dictionary """
        available_keys = list(self.keys())
        new_keys = pars.keys()
        mismatches = [key for key in new_keys if key not in available_keys]
        if len(mismatches):
            errormsg = f'Key(s) {mismatches} not found; available keys are {available_keys}'
            raise sc.KeyNotFoundError(errormsg)
        return

    def dict_update(self, *args, **kwargs):
        """ Redefine default dict.update(), since overwritten in this class; should not usually be used """
        super().update(*args, **kwargs)
        return

    def to_json(self, **kwargs):
        """ Convert to JSON representation """
        return sc.jsonify(dict(self), **kwargs)


class SimPars(Pars):
    """
    Create the parameters for the simulation. Typically, this function is used
    internally rather than called by the user; e.g. typical use would be to do
    sim = ss.Sim() and then inspect sim.pars, rather than calling this function
    directly.

    Args:
        label (str): The name of the simulation
        n_agents (int/float): The number of agents to run (default 10,000)
        total_pop (int/float): If provided, scale the agents to this effective population size
        pop_scale (float): If provided, use this agent-to-population scale factor (total_pop = n_agents*pop_scale)
        unit (str): The time unit for the simulation (default 'year'; other choices are 'day', 'week', 'month')
        start (float/str/date): The starting date for the simulation (default 2000); can be a year or date
        stop (float/str/date): If provided, the ending date for the simulation (if not provided, calculate from "dur")
        dur (int): How many timesteps to simulate, if "stop" is not provided (default 50)
        dt (float): The timestep, in units of "unit" (default 1.0)
        rand_seed (int): The overall random seed for the simulation (used to set module-specific random seeds)
        birth_rate (float): If provided, include births with this rate (per 1000 people per year)
        death_rate (float): If provided, include deaths with this rate (per 1000 people per year)
        use_aging (bool): Specify whether agents age (by default, agents age if and only if births and/or deaths are included)
        people (People): If provided, use a pre-existing People object rather than creating one (in which case n_agents will be ignored)
        networks (str/list/Module): The network module(s); can be a string, single module (i.e. Network), or list
        demographics (str/list/Module): As above
        diseases (str/list/Module): As above
        connectors (str/list/Module): As above
        interventions (str/list/Module): As above
        analyzers (str/list/Module): As above
        verbose (float): How much detail to print (1 = every timestep, 0.1 = every 10 timesteps, etc.)
    """
    def __init__(self, pars=None, create=True, **kwargs):

        # General parameters
        self.label   = '' # The label of the simulation
        self.verbose = ss.options.verbose # Whether or not to display information during the run -- options are 0 (silent), 0.1 (some; default), 1 (default), 2 (everything)

        # Population parameters
        self.n_agents  = 10e3 # Number of agents
        self.total_pop = None # If defined, used for calculating the scale factor
        self.pop_scale = None # How much to scale the population

        # Simulation parameters
        self.start     = None  # Start of the simulation (default 2020)
        self.stop      = None  # End of the simulation
        self.dur       = None  # Duration of time to run, if stop isn't specified (default 50 steps of self.unit)
        self.dt        = None  # Timestep
        self.rand_seed = 1     # Random seed; if None, don't reset

        # Demographic parameters
        self.birth_rate = None
        self.death_rate = None
        self.use_aging  = None # True if demographics, false otherwise

        # Modules
        self.people = None
        self._reset_modules()

        # Update with any supplied parameter values and generate things that need to be generated
        self.update(pars, create=create, **kwargs)
        return

    def _reset_modules(self):
        """ Initialize ndicts for module storage """
        self.custom        = ss.ndict()
        self.demographics  = ss.ndict()
        self.connectors    = ss.ndict()
        self.networks      = ss.ndict()
        self.interventions = ss.ndict()
        self.diseases      = ss.ndict()
        self.analyzers     = ss.ndict()
        self.modules       = ss.ndict()
        return

    def is_default(self, key):
        """ Check if the provided value matches the default """
        default_pars = SimPars() # Create a default SimPars object
        default_val = default_pars[key]
        current_val = self[key]
        match = (current_val == default_val) # Check if the value matches
        return match

    def validate(self):
        """ Call parameter validation methods """
        self.validate_sim_pars()
        self.validate_modules()
        return

    def validate_sim_pars(self):
        """ Validate each of the parameter values """
        self.validate_verbose()
        self.validate_agents()
        self.validate_total_pop()
        return

    def validate_verbose(self):
        """ Validate verbosity """
        if self.verbose == 'brief':
            self.verbose = -1
        if not sc.isnumber(self.verbose):  # pragma: no cover
            errormsg = f'Verbose argument should be either "brief", -1, or a float, not {type(self.verbose)} "{self.verbose}"'
            raise ValueError(errormsg)
        return

    def validate_agents(self):
        """ Check that n_agents is supplied and convert to an integer """
        if self.people is not None:
            len_people = len(self.people)
            if (len_people != self.n_agents) and not self.is_default('n_agents'):
                errormsg = f'Cannot set a custom value of n_agents ({self.n_agents}) that does not match a pre-created People object ({len_people}); use n_agents to create the People object instead'
                raise ValueError(errormsg)
            self.n_agents = len_people
        elif self.n_agents is not None:
            self.n_agents = int(self.n_agents)
        else:
            errormsg = 'Must supply n_agents, a people object, or a popdict'
            raise ValueError(errormsg)
        return

    def validate_total_pop(self):
        """ Ensure one but not both of total_pop and pop_scale are defined """
        if self.total_pop is not None:  # In future, if no pop_scale has been provided, try to get it from the location
            if self.pop_scale is not None:
                errormsg = f'You can define total_pop ({self.total_pop}) or pop_scale ({self.pop_scale}), but not both, since one is calculated from the other'
                raise ValueError(errormsg)
            total_pop = self.total_pop
        else:
            if self.pop_scale is not None:
                total_pop = self.pop_scale * self.n_agents
            else:
                total_pop = self.n_agents
        self.total_pop = total_pop
        if self.pop_scale is None:
            self.pop_scale = total_pop / self.n_agents
        return

    def validate_modules(self):
        """ Validate modules passed in pars"""

        # Do special validation on demographics (must be before modules are created)
        self.validate_demographics()

        # Get all modules into a consistent list format
        modmap = ss.module_map()
        for modkey in modmap.keys():
            if not isinstance(self[modkey], ss.ndict):
                self[modkey] = sc.tolist(self[modkey])

        # Convert any modules that are not already Module objects
        modules = self.convert_modules()
        self._reset_modules()
        for source, mod in modules:
            if source == 'modules':
                # Modules from 'modules=' get sorted by type
                for modkey, modclass in modmap.items():
                    if isinstance(mod, modclass):
                        self[modkey].append(mod)
                        break
                else:
                    self['custom'].append(mod)
            else:
                # Modules from specific containers stay where the user put them
                self[source].append(mod)

        # Do special validation on networks (must be after modules are created)
        self.validate_networks()
        return

    def validate_demographics(self):
        """ Validate demographics-related input parameters"""
        # Allow shortcut for default demographics
        demog = self.demographics
        if demog in [True, False, 1, 0]: # Allow turning it on or off
            self.demographics = sc.autolist()
            if demog: # ...turn it on with True or 1
                if self.birth_rate is None:
                    self.demographics += ss.Births()
                if self.death_rate is None:
                    self.demographics += ss.Deaths()

        # Allow users to add vital dynamics by entering birth_rate and death_rate parameters directly to the sim
        valid = isinstance(self.demographics, ss.ndict) and not len(self.demographics)
        if self.birth_rate is not None:
            if not valid:
                errormsg = 'You can only specify birth_rate together with (optionally) death_rate, not other demographics modules; add ss.Births() manually'
                raise ValueError(errormsg)
            births = ss.Births(birth_rate=self.birth_rate)
            self.demographics += births
        if self.death_rate is not None:
            if not valid:
                errormsg = 'You can only specify death_rate together with (optionally) birth_rate, not other demographics modules; add ss.Deaths() manually'
                raise ValueError(errormsg)
            background_deaths = ss.Deaths(death_rate=self.death_rate)
            self.demographics += background_deaths

        # Decide whether to use aging based on if demographics modules are present
        if self.use_aging is None:
            self.use_aging = True if self.demographics else False
        return

    def validate_networks(self):
        """ Validate networks """
        return

    def convert_modules(self):
        """
        Convert different types of representations for modules into a
        standardized object representation that can be parsed and used by
        a Sim object

        Returns a flat list of (source_key, ss.Module) tuples
        """

        out = []
        modmap = ss.module_map() # List of modules and parent module classes, e.g. ss.Disease
        modmap['modules'] = ss.Module # Enable passing in modules of any type via `Pars(modules=[...])`
        modmap['custom'] = ss.Module # Enable passing in modules of any type via `Pars(custom=[...])`
        supported_modules = ss.find_modules() # Each individual module class option, e.g. ss.SIR
        supported_modules['modules'] = sc.objdict({k:v for vv in supported_modules.values() for k,v in vv.items()})
        supported_modules['custom'] = supported_modules['modules']

        for modkey, ssmoddict in supported_modules.items():

            moddictkeys = ssmoddict.keys()
            moddictvals = ssmoddict.values()
            expected_cls = modmap[modkey]

            if isinstance(self[modkey], list):
                modlist = self[modkey]
            elif isinstance(self[modkey], ss.ndict):
                # If it is an ndict, then assume the keys specify names the values represent modules
                # They may already be module instances, in which case we will simply verify that the type
                # matches the expected type
                modlist = list(self[modkey].values())
            else:
                # If it is a dict with a single entry, then we wrap this in a list assuming that it is a dict
                # representation of a single module (as opposed to a dict containing modules)
                modlist = sc.promotetolist(self[modkey])

            for i,mod in enumerate(modlist):

                # Convert first from a string to a dict
                if isinstance(mod, str):
                    mod = dict(type=mod)

                # Convert from class to class instance (used for interventions and analyzers only)
                if isinstance(mod, type) and modkey in ['interventions', 'analyzers']:
                    mod = mod() # Call it to create a class instance

                # Now convert from a dict to a module
                if isinstance(mod, dict):

                    # Get the module type as a string
                    if 'type' in mod:
                        modtype = mod.pop('type')
                    else:
                        errormsg = f'When specifying a {modkey} module with a dict, one of the keys must be "type"; you supplied {mod}'
                        raise ValueError(errormsg)

                    # Get the module type as a class
                    if isinstance(modtype, str): # Usual case, a string, e.g. dict(type='sir', dur_inf=6)
                        modtype = modtype.lower() # Because our map is in lowercase
                        if modtype in moddictkeys:
                            modcls = ssmoddict[modtype]
                        else:
                            errormsg = f'Invalid module name "{modtype}" for "{modkey}"; must be one of {moddictkeys}'
                            raise TypeError(errormsg)
                    else: # Allow supplying directly as a class, e.g. dict(type=ss.SIR, dur_inf=6)
                        if modtype in moddictvals:
                            modcls = modtype
                        else:
                            errormsg = f'Invalid module class "{modtype}" for "{modkey}"; must be one of {moddictvals}'
                            raise TypeError(errormsg)

                    # Create the module and store it in the list
                    mod = modcls(**mod)

                # Special handling for interventions and analyzers: convert class and function to class instance
                if modkey in ['interventions', 'analyzers']:
                    if isinstance(mod, type) and issubclass(mod, expected_cls):
                        mod = mod()  # Convert from a class to an instance of a class
                    elif not isinstance(mod, ss.Module) and callable(mod):
                        mod = expected_cls.from_func(mod)

                # Convert plain modules from functions to actual modules
                if modkey == 'modules':
                    if not isinstance(mod, ss.Module) and callable(mod):
                        mod = ss.Module.from_func(mod)

                # Do final check
                if isinstance(expected_cls, type) and not isinstance(mod, (expected_cls, ss.Module)): # TEMP: check if this check still works?
                    errormsg = f'Was expecting {modkey} entry {i} to be class {expected_cls} or Module, but was {type(mod)} instead'
                    raise TypeError(errormsg)

                out.append((modkey, mod))

        return out



"""
Parent class for the integration loop.
"""
import time
import numpy as np
import pandas as pd
import sciris as sc
import starsim as ss
import matplotlib.pyplot as plt

# What classes are externally visible
__all__ = ['Loop']


#%% Loop class

class Loop:
    """
    Define the integration loop

    The Loop handles the order in which each function is called in the sim. The
    order is defined in `Loop.collect_funcs()`, which searches through the sim
    and collects all methods to call, in order, in the integration loop.

    Each type of module is called at a different time. Within each module type,
    they are called in the order listed. The default loop order is:

        1. sim:               start_step()     # Initialize the sim, including plotting progress
        2. all modules:       start_step()     # Initialize the modules, including the random number distribution
        3. sim.custom:        step()           # Run any custom modules
        4. sim.demographics:  step()           # Update the demographics, including adding new agents
        5. sim.diseases:      step_state()     # Update the disease states, e.g. exposed -> infected
        6. sim.connectors:    step()           # Run the connectors
        7. sim.networks:      step()           # Run the networks, including adding/removing edges
        8. sim.interventions: step()           # Run the interventions
        9. sim.diseases:      step()           # Run the diseases, including transmission
        10. people:           step_die()       # Figure out who died on this timestep
        11. people:           update_results() # Update basic state results
        12. all modules:      update_results() # Update any results
        13. sim.analyzers:    step()           # Run the analyzers
        14. all modules:      finish_step()    # Do any final tidying
        15. people:           finish_step()    # Clean up dead agents
        16. sim:              finish_step()    # Increment the timestep
    """
    def __init__(self, sim):
        self.sim = sim
        self.funcs = None
        self.func_list = []
        self.abs_tvecs = None
        self.plan = None
        self.index = 0 # The next function to execute
        self.cpu_time = [] # Store the CPU time of execution of each function
        self.df = None # User-friendly verison of the plan
        self.cpu_df = None # User-friendly time analysis
        self.initialized = False
        return

    def init(self):
        """ Parse the sim into the integration plan """
        self.collect_funcs()
        self.collect_abs_tvecs()
        self.make_plan()
        self.to_df()
        self.initialized = True
        return

    def __len__(self):
        if self.initialized:
            return len(self.plan)
        else:
            return 0 # Or None?

    def __iadd__(self, func):
        """ Allow functions to be added to the function list """
        parent = func.__self__
        func_name = func.__name__

        # Get the name if it's defined, the class otherwise; these must match abs_tvecs
        if isinstance(parent, ss.Module):
            module = parent.name
        elif isinstance(parent, ss.Sim):
            module = 'sim'
        else:
            module = parent.__class__.__name__.lower()

        # Create the row and append it to the function list
        func_path = f'{parent.__class__.__module__}.{func_name}'
        row = dict(
            func_order = len(self.funcs),
            module = module,
            func_name = func_name,
            func_path = func_path,
            func = func,
        )
        self.funcs.append(row)
        return self

    def __repr__(self):
        if self.initialized:
            arrs = list({len(arr) for arr in self.abs_tvecs.values()})
            if len(arrs) == 1: arrs = arrs[0] # If all are the same, just use that
            string = f'Loop(n={len(self)}, funcs={len(self.funcs)}, npts={arrs}, index={self.index})'
        else:
            string = 'Loop(initialized=False)'
        return string

    def disp(self):
        return sc.pr(self)

    def collect_funcs(self):
        """ Collect all the callable functions (methods) that comprise the step """

        # Run the simulation step first (updates the distributions)
        self.funcs = [] # Reset, just in case
        sim = self.sim

        # Collect the start_steps
        self += sim.start_step # Note special __iadd__() method above, which appends these to the funcs list
        for mod in sim.modules:
            self += mod.start_step

        # Update any nonspecific modules
        for mod in sim.custom():
            self += mod.step

        # Update demographic modules (create new agents from births/immigration, schedule non-disease deaths and emigration)
        for dem in sim.demographics():
            self += dem.step

        # Carry out autonomous state changes in the disease modules. This allows autonomous state changes/initializations
        # to be applied to newly created agents
        for disease in sim.diseases():
            if isinstance(disease, ss.Disease):
                self += disease.step_state

        # Update connectors
        for connector in sim.connectors():
            self += connector.step

        # Update networks - this takes place here in case autonomous state changes at this timestep
        # affect eligibility for contacts
        for network in sim.networks():
            self += network.step

        # Apply interventions - new changes to contacts will be visible and so the final networks can be customized by
        # interventions, by running them at this point
        for intv in sim.interventions():
            self += intv.step

        # Carry out autonomous state changes in the disease modules, including transmission (but excluding deaths)
        for disease in sim.diseases():
            self += disease.step

        # Update people who died -- calls disease.step_die() internally
        self += sim.people.step_die

        # Update results
        self += sim.people.update_results
        for mod in sim.modules:
            self += mod.update_results

        # Apply analyzers
        for ana in sim.analyzers():
            self += ana.step

        # Clean up dead agents, increment the time index, and perform other housekeeping tasks
        for mod in sim.modules:
            self += mod.finish_step
        self += sim.people.finish_step
        self += sim.finish_step

        return self.funcs

    def collect_abs_tvecs(self):
        """ Collect numerical time arrays for each module """
        self.abs_tvecs = sc.objdict()

        # Handle the sim and people first
        sim = self.sim
        for key in ['sim', sim.people.__class__.__name__.lower()]: # To handle subclassing of People -- TODO, make more elegant!
            self.abs_tvecs[key] = sim.t.tvec

        # Handle all other modules
        for mod in sim.modules:
            self.abs_tvecs[mod.name] = mod.t.tvec

        return self.abs_tvecs

    def make_plan(self):
        """ Combine the module ordering and the time vectors into the integration plan """
        # Assemble the list of dicts
        raw = []
        ti = -1
        for func_row in self.funcs:
            for t in self.abs_tvecs[func_row['module']]:
                row = func_row.copy()
                row['time'] = t # Add time column
                raw.append(row)

        # Turn it into a dataframe
        self.plan = sc.dataframe(raw)

        # Sort it by step_order, a combination of time and function order
        self.plan['ti'] = 0
        self.plan['label'] = self.plan.module + '.' + self.plan.func_name
        col_order = ['time', 'ti', 'func_order', 'func', 'label', 'module', 'func_name'] # Func in the middle to hide it
        self.plan = self.plan.sort_values(['time','func_order']).reset_index(drop=True)[col_order]

        # Calculate the sim time index (ti)
        start_step = 'sim.start_step'
        ti = -1
        ti_vals = []
        for i,label in enumerate(self.plan.label):
            if label == start_step:
                ti += 1
            ti_vals.append(ti)
        self.plan.loc[:, 'ti'] = ti_vals

        # Warn if any consecutive time values are close but not identical (likely a floating-point issue)
        eps = 1e-9
        try:
            diffs = np.diff(np.float64(self.plan.time)) # May be date, so convert to float
            small_diffs = diffs[(diffs > 0) & (diffs < eps)]
            if len(small_diffs):
                warnmsg = f'{len(small_diffs)} integration loop entries have near-identical times, indicating a floating-point issue:\n{small_diffs}\nCheck your time units across the sim and modules!'
                ss.warn(warnmsg)
        except Exception as E:
            warnmsg = f'Unable to calculate time deltas in integration plan; check that all types are compatible (e.g. float or ss.date): {E}'
            ss.warn(warnmsg)

        return

    def store_time(self):
        """ Store the current time in as high resolution as possible """
        self.cpu_time.append(time.perf_counter())
        return

    def run_one_step(self):
        """
        Take a single step, i.e. call a single function; only used for debugging purposes.

        Compare sim.run_one_step(), which runs a full timestep (which involves multiple function calls).
        """
        self._check_initialized()
        f = self.plan.func[self.index] # Get the next function
        f() # Call it
        self.index += 1 # Increment the time
        return

    def _check_initialized(self):
        """ Check that the Loop has been initialized """
        if not self.initialized:
            errormsg = 'Please initialize the loop (typically sim.init()) before calling this method.'
            raise RuntimeError(errormsg)
        return

    def run(self, until=None, verbose=None):
        """ Actually run the integration loop; usually called by sim.run() """
        self._check_initialized()

        # Convert e.g. '2020-01-01' to an actual date
        if isinstance(until, str):
            until = ss.date(until)

        # Loop over every function in the integration loop, e.g. disease.step()
        self.store_time()
        for f,label in zip(self.plan.func[self.index:], self.plan.label[self.index:]):
            if verbose:
                row = self.plan[self.index]
                print(f'Running t={row.time:n}, step={row.name}, {label}()')

            f() # Execute the function -- this is where all of Starsim happens!!

            # Tidy up
            self.index += 1 # Increment the count
            self.store_time()
            if until is not None and self.sim.now > until: # Terminate if asked to
                break

        self.to_df() # Store results as a dataframe
        return

    def insert(self, func, label=None, match_fn=None, before=False, verbose=True, die=True):
        """
        Insert a function into the loop plan at the specified location.

        The loop plan is a dataframe with columns including time (e.g. `date('2025-05-05')`),
        label (e.g. `'randomnet.step'`), module ('`randomnet'`), and function name (`'step'`).
        By default, this method will match the conditions in the plan based on
        the criteria specified.

        This functionality is similar to an analyzer or an intervention, but gives
        additional flexibility since can be inserted at (almost) any point in a sim.

        Note: the loop must be initialized (`sim.init()`) before you can call this.

        Args:
            func (func): the function to insert; must take a single argument, `sim`
            label (str): the label (module.name) of the function to match; see `sim.loop.plan.label.unique() for choices`
            match_fn (func): if supplied, use this function to perform the matching on the plan dataframe, returning a boolean array or list of indices of matching rows (see example below)
            before (bool): if true, insert the function before rather than after the match
            die (bool): whether to raise an exception if no matches found

        **Examples**:

            # Simple label matching with analyzer-like functionality
            def check_pop_size(sim):
                print(f'Population size is {len(sim.people)}')

            sim = ss.Sim(diseases='sir', networks='random', demographics=True)
            sim.init()
            sim.loop.insert(check_pop_size, label='people.finish_step')
            sim.run()

            # Function-based matching with intervention-like functionality
            def match_fn(plan):
                past_2010 = plan.time > ss.date(2010)
                is_step = (plan.label == 'sir.step') | (plan.label == 'randomnet.step')
                return past_2010 * is_step

            def update_betas(sim):
                if not sim.metadata.get('updated'):
                    print(f'Updating beta values on {sim.now}')
                    sim.diseases.sis.beta = 0.1
                    sim.networks.randomnet.edges.beta[:] = 0.5
                    sim.metadata.updated = True
                return

            sim = ss.Sim(diseases='sis', networks='random')
            sim.init()
            sim.loop.insert(update_betas, match_fn=match_fn, before=True)
            sim.run()
        """
        self._check_initialized()

        if label and match_fn:
            errormsg = "You can supply label or match, but not both; 'label' is equivalent to 'plan.label == label', please include this in your match function"
            raise ValueError(errormsg)

        if label:
            match_fn = lambda plan: plan.label == label

        # Compute the matches
        matches = match_fn(self.plan)
        if matches.dtype == bool:
            matches = sc.findinds(matches)

        # Perform the insertion in reverse order
        name = func.__name__
        sim_func = lambda: func(self.sim) # Construct a partial function
        for m in matches[::-1]:
            ind = m-1 if before else m
            current = self.plan[ind]
            row = dict(
                time = current.time,
                ti = current.ti,
                func_order = None,
                func = sim_func,
                label = name,
                module = None,
                func_name = name,
            )
            self.plan.insertrow(ind, row)

        return

    def to_df(self):
        """ Return a user-friendly version of the plan, omitting object columns """
        # Compute the main dataframe
        cols = ['time', 'ti', 'func_order', 'label', 'module', 'func_name']
        if self.plan is not None:
            df = self.plan[cols].copy() # Need to copy, otherwise it's messed up
        else:
            errormsg = f'Simulation "{self.sim}" needs to be initialized before exporting the Loop dataframe'
            raise RuntimeError(errormsg)
        times = np.diff(self.cpu_time)
        if len(times) == len(df):
            df['cpu_time'] = times
        else:
            df['cpu_time'] = np.nan
        self.df = df

        # Compute the CPU dataframe
        by_func = df.groupby('label')
        method = dict(func_order='first', module='first', func_name='first', cpu_time='sum')
        cdf = sc.dataframe(by_func.agg(method))
        cdf['percent'] = cdf.cpu_time / cdf.cpu_time.sum()*100
        cdf.insert(cdf.cols.index('cpu_time'), 'calls', by_func.size())
        cdf.sort_values('cpu_time', inplace=True, ascending=False)
        self.cpu_df = cdf
        return df

    def shrink(self):
        """ Shrink the size of the loop for saving to disk """
        shrunk = ss.utils.shrink()
        self.sim = shrunk
        self.funcs = shrunk
        self.plan = shrunk
        return

    def plot(self, simplify=False, max_len=100, fig_kw=None, plot_kw=None, scatter_kw=None):
        """
        Plot a diagram of all the events

        Args:
            simplify (bool): if True, skip update_results and finish_step events, which are automatically applied
            max_len (int): maximum number of entries to plot
            fig_kw (dict): passed to `plt.figure()`
            plot_kw (dict): passed to `plt.plot()`
            scatter_kw (dict): passed to `plt.scatter()`
        """

        # Assemble data
        df = self.to_df()
        if simplify:
            filter_out = ['update_results', 'finish_step']
            df = df[~df.func_name.isin(filter_out)]
        if max_len:
            df = df[:max_len]
        yticks = df.func_order.unique()
        ylabels = df.label.unique()
        x = df.time
        y = df.func_order

        # Convert module names to integers for plotting colors
        mod_int, _ = pd.factorize(df.module)
        colors = sc.gridcolors(np.unique(mod_int), asarray=True)

        # Do the plotting
        plot_kw = sc.mergedicts(dict(lw=2, alpha=0.2, c='k'), plot_kw)
        scatter_kw = sc.mergedicts(dict(s=200, alpha=0.6), scatter_kw)
        fig = plt.figure(**sc.mergedicts(fig_kw))
        plt.plot(x, y, **plot_kw)
        plt.scatter(x, y, c=colors[mod_int], **scatter_kw)
        plt.yticks(yticks, ylabels)
        plt.title(f'Integration plan ({len(df)} events)')
        plt.xlabel('Time since simulation start')
        plt.grid(True)
        sc.figlayout()
        sc.boxoff()
        return ss.return_fig(fig)

    def plot_cpu(self, bytime=True, max_entries=10, fig_kw=None, bar_kw=None):
        """
        Plot the CPU time spent on each event; visualization of Loop.cpu_df.

        Args:
            bytime (bool): if True, order events by total time rather than actual order
            fig_kw (dict): passed to `plt.figure()`
            bar_kw (dict): passed to `plt.bar()`
        """
        # Assemble data
        if self.cpu_df is None:
            self.to_df()
        df = self.cpu_df
        ylabels = df.index.values
        if bytime:
            y = np.arange(len(ylabels))
        else:
            y = df.func_order.values
        y = y[::-1] # Reverse order so plots from top to bottom

        x = df.cpu_time.values
        pcts = df.percent.values

        if x.max() < 1:
            x = x*1e3
            unit = 'ms'
        else:
            unit = 's'

        # Assemble labels
        for i in range(len(df)):
            timestr = sc.sigfig(x[i], 3) + f' {unit}'
            pctstr = sc.sigfig(pcts[i], 3) + '%'
            ylabels[i] += f'()\n{timestr}, {pctstr}'

        # Trim if needed
        if max_entries:
            x = x[:max_entries]
            y = y[:max_entries]
            ylabels = ylabels[:max_entries]

        # Do the plotting
        bar_kw = sc.mergedicts(bar_kw)
        fig = plt.figure(**sc.mergedicts(fig_kw))
        plt.barh(y, width=x, **bar_kw)
        plt.yticks(y, ylabels)
        plt.xlabel(f'CPU time ({unit})')
        plt.ylabel('Function call')
        plt.grid(True)
        sc.figlayout()
        sc.boxoff()
        return ss.return_fig(fig)

    def plot_step_order(self, which='default', max_len=500, plot_kw=None, scatter_kw=None, fig_kw=None, legend_kw=None):
        """
        Plot the order of the module steps across timesteps -- useful for debugging
        when using different time units.

        Note: generates a lot of data, best to debug with a small number of timesteps first!

        Args:
            which (dict): columns and values to filter to (default: {'func_name':'step'}; if None, do not filter)
            max_len (int): maximum number of entries to plot
            plot_kw (dict): passed to `plt.plot()`
            scatter_kw (dict): passed to `plt.scatter()`
            fig_kw (dict): passed to `plt.figure()`
            legend_kw (dict): passed to `plt.legend()`

        **Example**:

            sis = ss.SIS(dt=0.1)
            net = ss.RandomNet(dt=0.5)
            births = ss.Births(dt=1)
            sim = ss.Sim(dt=0.1, dur=5, diseases=sis, networks=net, demographics=births)
            sim.init()
            sim.loop.plot_step_order()
        """
        self._check_initialized()
        df = self.plan
        if which == 'default':
            which = dict(func_name='step')
        if which:
            for col,value in which.items():
                df = df[df[col] == value]
        if max_len and len(df) > max_len:
            print(f'Note: truncating from {len(df)} to {max_len} entries')
            df = df[:max_len]

        # Construct data
        unique = df.label.unique()
        n_unique = len(unique)
        colors = sc.gridcolors(n_unique)
        colormap = {k:v for k,v in zip(unique, colors)}
        d = sc.dictobj()
        for key in ['x', 'y', 'z', 'label']:
            d[key] = sc.autolist()
        for ti in df.ti.unique():
            this = df[df.ti==ti]
            d.x += list(range(len(this))) # Convert [0,0,0,...] to [0,1,2,...]
            d.y += list(this.func_order)
            d.z += list(this.ti)
            d.label += list(this.label)

        dd = sc.dataframe(d)

        fig = plt.figure(**sc.mergedicts(fig_kw))

        plot_kw = sc.mergedicts(dict(alpha=0.5, lw=2), plot_kw)
        scatter_kw = sc.mergedicts(dict(s=100, alpha=0.5), scatter_kw)
        ax = plt.axes(projection='3d')
        sc.plot3d(dd.x, dd.y, dd.z, ax=ax, **plot_kw)
        for label in unique:
            this = dd[dd.label==label]
            ax.scatter(this.x, this.y, this.z, color=colormap[label], label=label, **scatter_kw)
        ax.set_xlabel('Position within timestep')
        ax.set_ylabel('Original function order')
        ax.set_zlabel('Timestep')
        legend_kw = sc.mergedicts(dict(loc='upper left', bbox_to_anchor=(1.05, 1)), legend_kw)
        ax.legend(**legend_kw)
        return ss.return_fig(fig)

    def __deepcopy__(self, memo):
        """ A dataframe that has functions in it doesn't copy well; convert to a dict first """
        cls = self.__class__
        new = cls.__new__(cls)
        memo[id(self)] = new
        for k, v in vars(self).items():
            if k == 'plan' and isinstance(v, sc.dataframe):
                origdict = v.to_dict() # Convert to a dictionary
                newdict = sc.dcp(origdict, memo=memo, die=False) # Copy the dict
                newdf = sc.dataframe(newdict)
                setattr(new, k, newdf)
            else:
                setattr(new, k, sc.dcp(v, memo=memo, die=False)) # Regular deepcopy
        return new




"""
General module class -- base class for diseases, interventions, etc. Also
defines Analyzers and Connectors.
"""
import sys
import inspect
import functools as ft
import sciris as sc
import starsim as ss

__all__ = ['module_map', 'module_types', 'register_modules', 'find_modules', 'required', 'Base', 'Module']

module_args = ['name', 'label']  # Define allowable module arguments

custom_modules = []  # Allow the user to register custom modules


def module_map(key=None):
    """
    Map modules to standard types

    This is the source of truth about module types. Modules with types that do not
    map to anything in the module_map are treated as custom modules.
    """
    module_map = sc.objdict(
        demographics  = ss.Demographics,
        networks      = ss.Route, # NB, not ss.Network!
        diseases      = ss.Disease,
        connectors    = ss.Connector,
        interventions = ss.Intervention,
        analyzers     = ss.Analyzer,
    )
    return module_map if key is None else module_map[key]


def module_types():
    """ Return a list of known module types; based on `module_map()` """
    return list(module_map().keys())


def is_module(obj):
    """ Check whether something is a Starsim module """
    return isinstance(obj, type) and issubclass(obj, ss.Module)


def register_modules(*args):
    """
    Register custom modules with Starsim so they can be referred to by string.

    Note: "modules" here refers to Starsim modules. But this function registers
    Starsim "modules" (e.g. `ss.SIR`) that are found within Python "modules"
    (e.g. `starsim`).

    Args:
        args (list): the additional modules to register; can be either a module or a list of objects

    **Examples**:

        # Standard use case, register modules automatically
        import my_custom_disease_model as mcdm
        ss.register_modules(mcdm)
        ss.Sim(diseases='mydisease', networks='random').run() # This will work if mcdm.MyDisease() is defined

        # Manual usage
        my_modules = [mcdm.MyDisease, mcdm.MyNetwork]
        ss.register_modules(my_modules)
        ss.Sim(diseases='mydisease', networks='mynetwork').run()
    """
    for arg in args:
        custom_modules.append(arg)
    return


def find_modules(key=None, flat=False, verbose=False):
    """ Find all subclasses of Module present in Starsim, divided by type """
    modules = sc.objdict()
    modmap = module_map()
    attr_lists = [[ss, dir(ss)]] # Find all attributes in Starsim (note: does not parse user code; use register_modules() for that)
    for custom in custom_modules:
        if verbose: print(f'Loading {custom}...')
        if sc.ismodule(custom):
            attr_lists.append([custom, dir(custom)])
        else:
            attr_lists.append([None, sc.tolist(custom)])

    # Initialize dict
    for modkey in modmap.keys():
        modules[modkey] = sc.objdict()
    modules['modules'] = sc.objdict() # Catch-all for modules that do not match a standard type (e.g. user-created ss.Module classes)

    # Loop over all attributes
    for pymodule,attrs in attr_lists:
        if verbose: sc.heading(f'Processing {pymodule} ...')
        for attr in attrs: # Loop over each attribute (inefficient, but doesn't need to be optimized)
            if verbose: print(f'  Checking {attr}')

            # Handle modules or a list of objects being provided
            if pymodule is not None: # Main use case: pymodule=ss, attr='HIV'
                item = getattr(pymodule, attr)
                low_attr = attr.lower()
            else: # Alternate use case: pymodule=None, attr=HIV
                if is_module(attr): # It's a module, proceed
                    item = attr
                    low_attr = item.__name__.lower()
                else: # It's not a module, skip
                    item = None

            # Check whether it's a module
            ismodule = is_module(item)
            if ismodule:
                assigned = False
                for modkey, modtype in modmap.items(): # Loop over each module type
                    if modtype is not None and issubclass(item, modtype):
                        modules[modkey][low_attr] = item # It passes, so assign it to the dict
                        if modkey == 'networks' and low_attr.endswith('net'): # Also allow networks without 'net' suffix
                            altname = low_attr.removesuffix('net')
                            modules[modkey][altname] = item
                        assigned = True
                        if verbose: print(f'     Module {modkey}.{low_attr} added')
                        break # Exit the innermost loop
                if not assigned:
                    modules['modules'][low_attr] = item
                    if verbose: print(f'     Module modules.{low_attr} added')

    if flat:
        modules = sc.objdict({k:v for vv in modules.values() for k,v in vv.items()}) # Unpack the nested dict into a flat one
    return modules if key is None else modules[key]


def required(val=True):
    """
    Decorator to mark module methods as required.

    A common gotcha in Starsim is to forget to call super(), or to mistype a method
    name so it's never called. This decorator lets you mark methods (of Modules only)
    to be sure that they are called either on sim initialization or on sim run.

    Args:
        val (True/'disable'): by default, mark method as required; if set to 'disable', then disable method checking for parent classes as well (i.e. remove previous "required" calls)

    **Example**:

        class CustomSIS(ss.SIS):

            def step(self):
                super().step()
                self.custom_step() # Will raise an exception if this line is not here
                return

            @ss.required() # Mark this method as required on run
            def custom_step(self):
                pass
    """
    # Wrap the function
    def decorator(func):
        func._call_required = val  # Mark for collection

        @ft.wraps(func)
        def wrapper(self, *args, **kwargs):
            key = func.__qualname__
            self._call_required[key] += 1
            return func(self, *args, **kwargs)

        return wrapper
    return decorator


class Base:
    """
    The parent class for Sim and Module objects
    """
    def __bool__(self):
        """ Ensure that zero-length modules (e.g. networks) are still truthy """
        return True

    def __len__(self):
        """ The length of a module is the number of timepoints; see also len(sim) """
        try:    return self.t.npts
        except: return 0

    def __repr__(self):
        """ Equivalent to sc.prettyobj() """
        return sc.prepr(self)

    def disp(self, output=False, **kwargs):
        """ Display the full object """
        out = sc.prepr(self, **kwargs)
        if not output:
            print(out)
        else:
            return out

    @property
    def dt(self):
        """ Get the current module timestep """
        try:    return self.t.dt
        except: return None

    @property
    def ti(self):
        """ Get the current module timestep """
        try:    return self.t.ti
        except: return None

    @property
    def now(self):
        """ Shortcut to self.t.now() """
        try:    return self.t.now()
        except: return None

    @property
    def timevec(self):
        """ Shortcut to self.t.timevec """
        try:    return self.t.timevec
        except: return None

    def copy(self, die=True):
        """
        Perform a deep copy of the module/sim

        Args:
            die (bool): whether to raise an exception if copy fails (else, try a shallow copy)
        """
        out = sc.dcp(self, die=die)
        return out


class Module(Base):
    """
    The main base class for all Starsim modules: diseases, networks, interventions, etc.

    By convention, keyword arguments to modules are assigned to "_", which is an
    alias for None. These are then populated with defaults by `mod.define_pars()`,
    which are then updated with `mod.update_pars()`, which inspects the `__init__`
    function to get these arguments.

    Note that there is no *functional* difference between specifying arguments
    this way rather than simply via `**kwargs`, but having the arguments shown in the
    function signature can make it easier to read, and "_" is a convetion to indicate
    that the default is specified below.

    It is also of course OK to specify the actual values rather than "_"; however
    module arguments are often complex objects (e.g. `ss.bernoulli`) that can't
    be easily specified in the function signature.

    Finally, note that you can (and should) call `super().__init__()` with no arguments:
    the name, label, and time arguments get correctly updated via `self.update_pars()`.

    Args:
        name (str): a short, key-like name for the module (e.g. "randomnet")
        label (str): the full, human-readable name for the module (e.g. "Random network")
        kwargs (dict): passed to `ss.Timeline()` (e.g. start, stop, unit, dt)

    **Example**:

        class SIR(ss.Module):
            def __init__(self, pars=_, beta=_, init_prev=_, p_death=_, **kwargs):
                super().__init__() # Call this first with no arguments
                self.define_pars( # Then define the parameters, including their default value
                    beta = ss.peryear(0.1),
                    init_prev = ss.bernoulli(p=0.01),
                    p_death = ss.bernoulli(p=0.3),
                )
                self.update_pars(pars, **kwargs) # Update with any user-supplied parameters, and raise an exception if trying to set a parameter that wasn't defined in define_pars()
                return
    """
    def __init__(self, name=None, label=None, **kwargs):
        # Housekeeping
        self._locked_attrs = ['pars', 't', 'sim', 'dists', 'results'] # Define key attributes that shouldn't be overwritten
        self._collect_required() # First, collect methods marked as required on creation

        # Handle parameters
        self.pars = ss.Pars() # Usually populated via self.define_pars()
        self.set_metadata(name, label) # Usually reset as part of self.update_pars()
        self.t = ss.Timeline(**kwargs, name=self.name)

        # Properties to be added by init_pre()
        self.sim = None
        self.dists = None # Turned into a Dists object by sim.init_dists() if this module has dists
        self.results = ss.Results(self.label)

        # Finish initialization
        self.pre_initialized = False
        self.initialized = False
        self.finalized = False
        self._lock_attrs = True # Prevent key attributes from being overwritten directly
        self._auto_states = [] # Store automatic states (State objects) added via define_states()
        return

    def __call__(self, *args, **kwargs):
        """ Allow modules to be called like functions """
        return self.step(*args, **kwargs)

    def __getitem__(self, key):
        """ Allow modules to act like dictionaries """
        return getattr(self, key)

    def __setattr__(self, attr, value):
        """ Don't allow locked attributes to be overwritten """
        if getattr(self, '_lock_attrs', False) and attr in self._locked_attrs:
            if value is getattr(self, attr, None): # Allow assigning the same object back - this happens automatically with some in-place operators (e.g., +=)
                return
            errormsg = f'Cannot modify attribute "{attr}"; locked attributes are {sc.strjoin(self._locked_attrs)}.\n'
            errormsg += 'If you really mean to do this, use module.setattribute() or set module._lock_attrs = False'
            try:
                if isinstance(getattr(self, attr),ss.Arr) and len(getattr(self, attr)) == len(value):
                    errormsg += f'\nOften this error can be resolved by replacing `self.{attr} = x` with `self.{attr}[:] = x`'
            except:
                pass
            raise AttributeError(errormsg)
        else:
            super().__setattr__(attr, value)
        return

    def setattribute(self, attr, value):
        """ Method for setting an attribute that does not perform checking against immutable attributes """
        return super().__setattr__(attr, value)

    def brief(self, output=False):
        """ Show a brief representation of the module; used by __repr__ """
        name = self.name # e.g. 'sir'
        label = self.label # e.g. 'My SIR'
        class_name = self.__class__.__name__ # e.g. 'SIR'
        class_str = f':{class_name}' if name != class_name.lower() else '' # e.g. 'SIR'
        label_str = f'"{label}"; ' if (label.lower() != name and label != class_name) else ''
        pars_str = '[' + sc.strjoin(self.pars.keys()) + ']' if len(self.pars) else 'None'
        states_str = '[' + sc.strjoin(self.state_dict.keys()) + ']' if len(self.state_dict) else 'None'
        out = f'{name}{class_str}({label_str}pars={pars_str}; states={states_str})'
        return out if output else print(out)

    def __repr__(self):
        """ Show the object including parameters and states """
        try: # Default representation
            return self.brief(output=True)
        except: # If for any reason that fails, return the full repr
            return sc.prepr(self)

    @property
    def _debug_name(self):
        """ The module name and class shown as a string, used in debugging/error messages """
        out = f'"{self.name}" ({type(self)}'
        return out

    def _reconcile(self, key, value=None, default=None):
        """ Reconcile module attributes, parameters, and input arguments """
        parval = self.pars.get(key)
        attrval = getattr(self, key, parval)
        val = sc.ifelse(value, attrval, default)
        return val

    def _collect_required(self):
        """ Collect all methods marked as required """
        reqs = {}
        disabled = []
        for cls in inspect.getmro(type(self)):
            for attr,method in cls.__dict__.items():
                req = getattr(method, '_call_required', False)
                if req:
                    valid = [True, 'disable']
                    if req not in valid:
                        errormsg = f'ss.require() must be True or "disable", not "{req}"'
                        raise ValueError(errormsg)
                    elif req == 'disable':
                        disabled.append(attr)
                    key = f'{cls.__qualname__}.{attr}'
                    reqs[key] = 0 # Meaning it's been called 0 times

        # Collate and manually increment any that are set to be disabled
        self._call_required = reqs
        if disabled:
            for k in self._call_required.keys():
                if k.split('.')[-1] in disabled: # Look for matches for disabled, omitting the class name
                    self._call_required[k] += 1 # Manually increment the call to pass checking
        return required

    def check_method_calls(self):
        """
        Check if any required methods were not called.

        Typically called automatically by `sim.run()`.
        """
        missing = [key for key, called in self._call_required.items() if not called]
        return missing

    @required()
    def set_metadata(self, name=None, label=None):
        """ Set metadata for the module """
        # Validation
        for key,val in dict(name=name, label=label).items():
            if val is not None:
                if not isinstance(val, str):
                    errormsg = f'Invalid value for {key}: must be str, not {type(val)}: {val}'
                    raise TypeError(errormsg)

        # Set values for name and label
        cls_name = self.__class__.__name__
        cls_lower = cls_name.lower()
        self.name = self._reconcile('name', name, cls_lower)
        default_label = self.name if self.name != cls_lower else cls_name
        self.label = self._reconcile('label', label, default_label)
        return

    @classmethod
    def create(cls, name, *args, **kwargs):
        """
        Create a module instance by name

        Args:
            name (str): A string with the name of the module class in lower case, e.g. 'sir'
        """
        for subcls in ss.all_subclasses(cls):
            if subcls.__name__.lower() == name:
                return subcls(*args, **kwargs)
        else:
            raise KeyError(f'Module "{name}" did not match any known Starsim modules')

    @classmethod
    def from_func(cls, func):
        """ Create an module from a function """
        def step(mod): # TODO: see if this can be done more simply
            return mod.func(mod.sim)
        name = func.__name__
        mod = cls(name=name)
        mod.func = func
        mod.step = ft.partial(step, mod)
        mod.step.__name__ = name # Manually add these in as for a regular class method
        mod.step.__self__ = mod
        return mod

    def define_pars(self, inherit=True, **kwargs): # TODO: think if inherit should default to true or false
        """
        Create or merge Pars objects

        Note: this method also automatically pulls in keyword arguments from the
        calling function (which is almost always a module's `__init__()` method)
        """
        if inherit: # Merge with existing
            self.pars.update(**kwargs, create=True)
        else: # Or overwrite
            self.pars = ss.Pars(**kwargs)
        return self.pars

    # Warning: do not try to use a decorator with this function, that will break argument passing!
    def update_pars(self, pars=None, **kwargs):
        """
        Pull out recognized parameters, returning the rest
        """
        # Merge pars and kwargs
        pars = sc.mergedicts(pars, kwargs)

        # Inspect the parent frame and pull out any arguments
        frame = sys._getframe(1)  # Go back 1 frame (to __init__, most likely)
        if frame.f_code.co_name == '__init__': # If it's not being called from init, don't do this
            _, _, _, kw = inspect.getargvalues(frame) # Get the values provided
            for k,v in kw.items():
                if k in self.pars and v is not None:
                    pars[k] = v

        # Update matching module parameters
        matches = {}
        for key in list(pars.keys()): # Need to cast to list to avoid "dict changed during iteration"
            if key in self.pars:
                matches[key] = pars.pop(key)
        self.pars.update(matches)

        # Update module attributes
        metadata = {key:pars.get(key, self.pars.get(key)) for key in module_args}
        timepars = {key:pars.get(key, self.pars.get(key)) for key in ss.Timeline.time_args}
        self.set_metadata(**metadata)
        self.t.update(**timepars)

        # Should be no remaining pars
        remaining = set(pars.keys()) - set(module_args) - set(ss.Timeline.time_args)
        if len(remaining):
            errormsg = f'{len(pars)} unrecognized arguments for {self.name}: {sc.strjoin(remaining)}'
            raise ValueError(errormsg)
        return

    def define_states(self, *args, check=True, reset=False, lock=True):
        """
        Define states of the module with the same attribute name as the state

        In addition to registering the state with the module by attribute, it adds
        it to `mod._all_states`, which is used by `mod.state_list` and `mod.state_dict`.

        Args:
            args (states): list of states to add
            check (bool): whether to check that the object being added is a state, and that it's not already present
            reset (bool): whether to reset the list of module states and use only the ones provided
            lock (bool): if True, prevent states from being
        """
        # Optionally reset the states (note: does not remove them from the people object or others if already added); see example in ss.SIR()
        if reset:
            for state in self.state_list:
                attr = state.name
                delattr(self, attr)
                if attr in self._locked_attrs:
                    self._locked_attrs.remove(attr)
            self._auto_states = []

        # If we're not checking, don't lock the attrs
        if not check:
            orig_lock_attrs = self._lock_attrs
            self._lock_attrs = False

        # Add the new states
        for arg in args:
            if isinstance(arg, (list, tuple)):
                state = ss.BoolState(*arg)
            elif isinstance(arg, dict):
                state = ss.BoolState(**arg)
            else:
                state = arg

            if check:
                assert isinstance(state, ss.Arr), f'Could not add {state}: not an Arr object'

            # Add the state to the module
            attr = state.name
            if check and hasattr(self, attr):
                present = [s.name for s in self.state_list]
                new = [s.name for s in args]
                errormsg = f'Cannot add "{attr}" to {self._debug_name} since already present in module.\n'
                errormsg += 'Did you mean to use define_states(reset=True) (skip inherited states) or define_states(check=False) (skip this check)?\n'
                errormsg += f'States already in module:\n{present}\n'
                errormsg += f'New states being added:\n{new}\n'
                errormsg += f'Conflicting states:\n{set(present) & set(new)}\n'
                raise AttributeError(errormsg)
            setattr(self, attr, state)
            if lock:
                self._locked_attrs.append(attr)

            # Add it to the list of auto states, if needed
            if isinstance(state, ss.BoolState):
                self._auto_states.append(state)

        # Reset the lock state
        if not check:
            self._lock_attrs = orig_lock_attrs
        return

    def define_results(self, *args, check=True):
        """ Add results to the module """
        for arg in args:
            if isinstance(arg, (list, tuple)):
                result = ss.Result(*arg)
            elif isinstance(arg, dict):
                result = ss.Result(**arg)
            else:
                result = arg

            # Update with module information
            result.update(module=self.label, shape=self.t.npts, timevec=self.t.timevec)

            # Add the result to the dict of results; does automatic checking
            if result.name in self.results and result.name.startswith('n_') and result.name[2:] in {state.name for state in self.auto_state_list}:
                msg = f'"{self.name}": Another result named "{result.name}" already exists because a result was automatically created for the BoolState "{result.name[2:]}".'
                raise ValueError(msg)
            self.results += result
        return

    @required()
    def init_pre(self, sim, force=False):
        """
        Perform initialization steps

        This method is called once, as part of initializing a Sim. Note: after
        initialization, initialized=False until init_vals() is called (which is after
        distributions are initialized).
        """
        if force or not self.pre_initialized:
            self.setattribute('sim', sim) # Link back to the sim object
            ss.link_dists(self, sim, skip=[ss.Sim, ss.Module]) # Link the distributions to sim and module, skipping any nested sim or module instances
            self.t.init(sim=self.sim) # Initialize time vector
            self.link_rates() # Add module dt to the timepars
            sim.pars[self.name] = self.pars
            sim.results[self.name] = self.results
            sim.people.add_module(self) # Connect the states to the people
            self.init_results()
            self.pre_initialized = True
        return

    @required()
    def link_rates(self, force=False):
        """ Find all time parameters in the module and link them to the module's dt """
        rates = sc.search(self, type=ss.Rate, skip=dict(keys=['sim', 'module'])) # Should it be self or self.pars?

        # Initialize them with the parent module
        for rate in rates.values():
            if force or rate.default_dur is None:
                rate.set_default_dur(self.t.dt)
        return

    @required()
    def init_results(self):
        """
        Initialize results output; called during `init_pre()`

        By default, modules all report on counts for any explicitly defined "States", e.g. if
        a disease contains an `ss.BoolState` called 'susceptible' it will automatically contain a
        Result for 'n_susceptible'. For identical behavior that does not automatically
        generate results, use `ss.BoolArr` instead of `ss.BoolState`.
        """
        self.results.timevec = self.t.timevec # Store the timevec in the results for plotting; not a Result so don't use ss.ndict.append()
        results = sc.autolist()
        for state in self.auto_state_list:
            results += ss.Result(f'n_{state.name}', dtype=int, scale=True, label=state.label)
        self.define_results(*results)
        return

    @required()
    def init_post(self):
        """ Initialize the values of the states; the last step of initialization """
        for state in self.state_list:
            if not state.initialized:
                state.init_vals()
        self.initialized = True
        return

    def init_mock(self, n_agents=100, dur=10):
        """ Initialize with a mock simulation -- for debugging purposes only """
        sim = ss.mock_sim(n_agents=n_agents, dur=dur)
        self.init_pre(sim)
        for state in self.state_list: # Manually link the people
            state.people = self.sim.people
        obj = [self.pars, self.state_list]
        dists = sc.search(obj, type=ss.Dist)
        for i,dist in dists.enumvals():
            dist.trace = f'dist_{i}'
        ss.link_dists(obj=obj, sim=sim, module=self, skip=[ss.Sim, ss.Module], init=True)
        self.init_post()
        return self

    @property
    def state_list(self):
        """
        Return a flat list of all states (`ss.Arr` objects)

        The base class returns all states that are contained in top-level attributes
        of the Module. If a Module stores states in a non-standard location (e.g.,
        within a list of states, or otherwise in some other nested structure - perhaps
        due to supporting features like multiple genotypes) then the Module should
        overload this attribute to ensure that all states appear in here.
        """
        out = [val for val in self.__dict__.values() if isinstance(val, ss.Arr)]
        return out

    @property
    def state_dict(self):
        """
        Return a flat dictionary (objdict) of all states

        Will raise an exception if multiple states have the same name.
        """
        return ss.utils.nlist_to_dict(self.state_list)

    @property
    def auto_state_list(self):
        """
        List of "automatic" states with boolean type (`ss.BoolState`) that were added
        via `define_states()`

        For diseases, these states typically represent attributes like 'susceptible',
        'infectious', 'diagnosed' etc. These variables automatically generate results
        like n_susceptible, n_infectious, etc. For a list of all states, see
        `state_list`.
        """
        return self._auto_states[:]

    def match_time_inds(self, inds=None):
         """ Find the nearest matching sim time indices for the current module """
         self_tvec = self.t.yearvec
         sim_tvec = self.sim.t.yearvec
         if len(self_tvec) == len(sim_tvec): # Shortcut to avoid doing matching
             return Ellipsis if inds is None else inds
         else:
             out = sc.findnearest(sim_tvec, self_tvec)
             return out

    @required()
    def start_step(self):
        """ Tasks to perform at the beginning of the step """
        if self.finalized:
            errormsg = f'The module {self._debug_name} has already been run. Did you mean to copy it before running it?'
            raise RuntimeError(errormsg)
        if self.dists is not None: # Will be None if no distributions are defined
            self.dists.jump_dt() # Advance random number generators forward for calls on this step
        return

    def step(self):
        """ Define how the module updates over time -- the key part of Starsim!! """
        errormsg = f'The module {self._debug_name} does not define a "step" method. This is usually a bug, but use "def step(self): pass" if this is intentional.'
        ss.warn(errormsg)
        return

    @required()
    def finish_step(self):
        """ Define what should happen at the end of the step; at minimum, increment ti """
        self.t.ti += 1
        return

    def update_results(self):
        """
        Update results; by default, compute counts of each state at each point in time

        This function is executed after transmission in all modules has been resolved.
        This allows result updates at this point to capture outcomes dependent on multiple
        modules, where relevant.
        """
        for state in self.auto_state_list:
            self.results[f'n_{state.name}'][self.ti] = state.sum()
        return

    @required()
    def finalize(self):
        """ Perform any final operations, such as removing unneeded data """
        # Update the time index (since otherwise pointing to a timepoint that doesn't exist)
        self.t.ti -= 1

        self.finalize_results()
        self.finalized = True
        return

    @required()
    def finalize_results(self): # TODO: this is confusing, needs to be not redefined by the user, or called *after* a custom finalize_results()
        """ Finalize results """
        # Scale results
        for reskey, res in self.results.items():
            if isinstance(res, ss.Result) and res.scale:
                self.results[reskey] = self.results[reskey]*self.sim.pars.pop_scale
        return

    def to_json(self):
        """ Export to a JSON-compatible format """
        out = sc.objdict()
        out.type = self.__class__.__name__
        out.name = self.name
        out.label = self.label
        out.pars = self.pars.to_json()
        return out

    def shrink(self):
        """ Shrink the size of the module for saving to disk """
        shrunk = ss.utils.shrink()
        self.setattribute('sim', shrunk) # Use setattribute since locked otherwise
        self.setattribute('dists', shrunk)
        for state in self.state_list:
            with sc.tryexcept():
                state.people = shrunk
                state.raw = shrunk
        return

    def plot(self):
        """ Plot all results in the module """
        with sc.options.with_style('fancy'):
            flat = sc.flattendict(self.results, sep=': ')
            timevec = self.t.timevec
            fig, axs = sc.getrowscols(len(flat), make=True)
            for ax, (k, v) in zip(axs.flatten(), flat.items()):
                ax.plot(timevec, v)
                ax.set_title(k)
                ax.set_xlabel('Year')
        return ss.return_fig(fig)


"""
Run syphilis
"""

# %% Imports and settings
import numpy as np
import pandas as pd
import sciris as sc
import starsim as ss
import starsim_examples as sse
import matplotlib.pyplot as plt

quick_run = True
datadir = ss.root / 'tests/test_data'


def make_syph_sim(dt=1, n_agents=500):
    """ Make a sim with syphilis - used by several subsequent tests """
    syph = sse.Syphilis()
    syph.pars.beta = dict(mf=[0.25, 0.15], maternal=[0.99, 0])
    syph.pars.init_prev = ss.bernoulli(p=0.1)

    # Make demographic modules
    pregnancy = ss.Pregnancy(fertility_rate=pd.read_csv(datadir/'nigeria_asfr.csv'))
    death = ss.Deaths(death_rate=pd.read_csv(datadir/'nigeria_deaths.csv'), rate_units=1)

    # Make people and networks
    ppl = ss.People(n_agents, age_data=datadir/'nigeria_age.csv')

    # Marital
    mf = ss.MFNet(duration=1/24)
    maternal = ss.MaternalNet()

    sim_kwargs = dict(
        dt=dt,
        total_pop=93963392,
        start=1990,
        dur=40,
        people=ppl,
        diseases=syph,
        networks=ss.ndict(mf, maternal),
        demographics=[pregnancy, death],
    )

    return sim_kwargs


class check_states(ss.Analyzer):
    """ Analyzer to check consistency of states """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = 'check_states'
        self.okay = True
        return

    def step(self):
        """
        Checks states that should be mutually exlusive and collectively exhaustive
        """
        sim = self.sim
        sppl = sim.diseases.syphilis

        # Infection states: people must be exactly one of these
        s1a = (sppl.susceptible | sppl.exposed | sppl.primary | sppl.secondary | sppl.latent_temp | sppl.latent_long | sppl.tertiary | sppl.congenital).all()
        s1b = (sppl.naive | sppl.sus_not_naive | sppl.exposed | sppl.primary | sppl.secondary | sppl.latent_temp | sppl.latent_long | sppl.tertiary | sppl.congenital).all()
        if not (s1a & s1b):
            raise ValueError('States should be collectively exhaustive but are not.')
        s2a = ~(sppl.susceptible & sppl.exposed & sppl.primary & sppl.secondary & sppl.latent_temp & sppl.latent_long & sppl.tertiary & sppl.congenital).any()
        s2b = ~(sppl.naive & sppl.sus_not_naive & sppl.exposed & sppl.primary & sppl.secondary & sppl.latent_temp & sppl.latent_long & sppl.tertiary & sppl.congenital).any()
        if not (s2a & s2b):
            raise ValueError('States should be mutually exclusive but are not.')
        s3 = ~(sppl.susceptible & sppl.infected).any()
        if not s3:
            raise ValueError('States S and I should be mutually exclusive but are not.')

        checkall = np.array([s1a, s1b, s2a, s2b, s3])
        if not checkall.all():
            self.okay = False
        return


@sc.timer()
def test_syph(dt=1, n_agents=500, do_plot=False):

    sim_kwargs = make_syph_sim(dt=dt, n_agents=n_agents)
    sim = ss.Sim(analyzers=check_states(), **sim_kwargs)
    sim.run()

    # Check plots
    burnin = 0

    if do_plot:
        tvec = sim.timevec[burnin:]
        res = sim.results.syphilis
        fig, ax = plt.subplots(2, 2)
        ax = ax.ravel()
        ax[0].stackplot(
            tvec,
            # res.n_susceptible[burnin:],
            res.n_congenital[burnin:],
            res.n_exposed[burnin:],
            res.n_primary[burnin:],
            res.n_secondary[burnin:],
            (res.n_latent_temp[burnin:]+res.n_latent_long[burnin:]),
            res.n_tertiary[burnin:],
        )
        ax[0].legend(['Congenital', 'Exposed', 'Primary', 'Secondary', 'Latent', 'Tertiary'], loc='lower right')

        ax[1].plot(tvec, res.prevalence[burnin:])
        ax[1].set_title('Syphilis prevalence')

        ax[2].plot(tvec, sim.results.n_alive[burnin:])
        ax[2].set_title('Population')

        ax[3].plot(tvec, res.new_infections[burnin:])
        ax[3].set_title('New infections')

        fig.tight_layout()
        plt.show()

    return sim


@sc.timer()
def test_syph_intvs(dt=1, n_agents=500, do_plot=False):

    # Interventions
    # screen_eligible = lambda sim: sim.demographics.pregnancy.pregnant
    screen_eligible = lambda sim: sim.networks.mfnet.active(sim.people)
    syph_screening = sse.syph_screening(
        product='rpr',
        prob=0.99,
        eligibility=screen_eligible,
        start_year=2020,
        label='syph_screening',
    )

    treat_eligible = lambda sim: ss.uids(sim.interventions['syph_screening'].outcomes['positive'])
    bpg = sse.syph_treatment(
        prob=0.9,
        product='bpg',
        eligibility=treat_eligible,
        label='bpg'
    )

    sim_kwargs1 = make_syph_sim(dt=dt, n_agents=n_agents)
    sim_intv = ss.Sim(analyzers=[check_states()], interventions=[syph_screening, bpg], **sim_kwargs1)
    sim_intv.run()

    # Check plots
    if do_plot:
        # Run baseline
        sim_kwargs0 = make_syph_sim(dt=dt, n_agents=n_agents)
        sim_base = ss.Sim(**sim_kwargs0)
        sim_base.run()

        burnin = 10
        syph_b = sim_base.diseases.syphilis
        syph_i = sim_intv.diseases.syphilis

        plt.figure()
        plt.plot(syph_b.timevec[burnin:], syph_b.results.prevalence[burnin:], label='Baseline')
        plt.plot(syph_i.timevec[burnin:], syph_i.results.prevalence[burnin:], label='S&T')
        plt.axvline(x=ss.date(2020), color='k', ls='--')
        plt.title('Syphilis prevalence')
        plt.legend()
        plt.show()

        return sim_base, sim_intv

    else:
        return sim_intv


if __name__ == '__main__':
    T = sc.timer()
    do_plot = True

    dt = [ss.months(1), ss.years(1)][quick_run]
    n_agents = [20e3, 500][quick_run]

    sim = test_syph(dt=dt, n_agents=n_agents, do_plot=do_plot)
    sim_base, sim_intv = test_syph_intvs(dt=dt, n_agents=n_agents, do_plot=True)

    T.toc()




"""
Test different time units and timesteps
"""
import numpy as np
import sciris as sc
import starsim as ss
import pytest

# ss.options.warnings = 'error' # For additional debugging


@sc.timer()
def test_ratio():
    sc.heading('Test dur/datedur time ratio calculation')

    assert ss.years(1) / ss.years(1) == 1
    assert ss.days(1) / ss.days(0.1) == 10
    assert ss.weeks(0.5) / ss.weeks(5) == 0.1

    assert ss.datedur(years=1) / ss.datedur(days=1) == 365
    assert np.isclose(ss.datedur(years=1) / ss.datedur(weeks=1) * 7, 365)
    assert ss.datedur(years=1) / ss.datedur(months=1) == 12

    return


@sc.timer()
def test_classes():
    sc.heading('Test behavior of dur() and rate()')

    # Test duration dt
    d1 = ss.years(2)
    d2 = ss.years(3)
    d3 = ss.years(2/0.1)
    d4 = ss.years(3/0.2)

    assert d1 + d2 == 2+3
    assert d3 + d4 == 2/0.1 + 3/0.2
    assert d3 * 2 == 2/0.1*2
    assert d3 / 2 == 2/0.1/2

    # Test rate dt
    r1 = ss.freqperyear(2)
    r2 = ss.freqperyear(3)
    r3 = ss.freqperyear(2/0.1)
    r4 = ss.freqperyear(3/0.2)

    assert r1 + r2 == ss.freqperyear(5)
    assert r3 + r4 == ss.freqperyear(20+15)
    assert r3 * 2 == ss.freqperyear(4/0.1)
    assert r4 / 2 == ss.freqperyear(1.5/0.2)

    # Test duration units
    d5 = ss.datedur(years=2)
    d6 = ss.datedur(days=5)
    assert (d5 + d6).years == 2 + 5/365
    assert (d5 + d6)/ss.datedur(days=1) == 365*2+5

    # Test rate units
    rval = 0.7
    r5 = ss.freq(rval, ss.week)
    assert np.isclose(r5*ss.days(1), rval/7) # These should be close, but not match exactly
    assert np.isclose(r5*ss.weeks(0.1), rval/10)
    assert r5*ss.days(1) == rval * ss.days(1) / ss.weeks(1) # These should match exactly
    assert r5*ss.weeks(0.1) == rval * ss.weeks(0.1) / ss.weeks(1)

    # Test prob
    tpval = 0.1
    tp0 = ss.probperyear(tpval)
    assert np.isclose(tp0*ss.years(1), tpval), 'Multiplication by the base denominator should not change the value'
    assert np.isclose(tp0*ss.years(0.5), tpval/2, rtol=0.1) # These should be close, but not match exactly
    assert np.isclose(tp0*ss.years(2), tpval*2, rtol=0.1)
    assert tp0*ss.years(0.5) == 1 - np.exp(np.log(1-0.1) * ss.years(0.5)/ss.years(1)) # These should be close, but not match exactly
    assert tp0*ss.years(2) == 1 - np.exp(np.log(1-0.1) * ss.years(2)/ss.years(1)) # These should be close, but not match exactly

    return d3, d4, r3, r4, tp0



@sc.timer()
def test_callable_dists():
    sc.heading('Testing callable distributions')
    def loc(module, sim, uids):
        return np.arange(uids)
    module = ss.mock_module(dt=ss.day)
    d = ss.normal(loc, 2, unit='days', module=module, strict=False)
    d.init()
    d.rvs(10)
    return d


@sc.timer()
def test_syntax():
    """ Verify that a range of supported operations run without raising an error """
    sc.heading('Testing full syntax')

    assert float(ss.date(1500))==1500
    assert np.isclose(float(ss.date(1500.1)), 1500.1) # Not exactly equal, but very close
    assert np.all((ss.years(1)*np.arange(5)) == (np.arange(5)*ss.years(1)))
    assert np.isclose(ss.datedur(weeks=1)/ss.datedur(days=1), 7) # TODO: would be nice if this were exact, but maybe impossible
    assert np.isclose(float(ss.datedur(weeks=1) - ss.datedur(days=1)), 6/365)
    assert np.isclose((ss.date(2050)-ss.date(2020)).years, ss.years(30).years, rtol=1/365) # Not exact due to leap years
    assert np.isclose((ss.freqperweek(1)+ss.freqperday(1)).value, ss.freqperweek(8).value) # CKTODO: would be nice if this were exact

    assert ss.date('2020-01-01') + ss.datedur(weeks=52) == ss.date('2020-12-30') # Should give us 30th December 2020
    assert ss.date('2020-01-01') + 52*ss.datedur(weeks=1) == ss.date('2020-12-30')# Should give us 30th December 2020
    assert ss.date('2020-01-01') + 52*ss.years(1/52) == ss.date('2021-01-01') # Should give us 1st Jan 2021
    assert ss.date('2020-01-01') + ss.datedur(years=1) == ss.date('2021-01-01') # Should give us 1st Jan 2021

    # Operations on date vectors
    ss.date.arange(2020,2030)+ss.years(1) # add years to date array
    ss.date.arange(2020,2030)+ss.datedur(years=1) # add datedur to date array

    # Construction of various duration ranges and addition with durations and dates
    ss.dur.arange(ss.years(0), ss.years(10),ss.years(1)) + ss.years(1)
    ss.dur.arange(ss.years(0), ss.years(10), ss.datedur(years=1)) + ss.years(1)
    ss.dur.arange(ss.years(0), ss.datedur(years=10), ss.datedur(years=1)) + ss.years(1)

    # Datedur calculations -- these are not all currently working
    # ss.dur.arange(ss.datedur(years=0), ss.datedur(years=10), ss.datedur(years=1)) + ss.years(1)
    # ss.dur.arange(ss.datedur(years=0), ss.datedur(years=10), ss.datedur(years=1)) + ss.datedur(years=1)
    # ss.dur.arange(ss.datedur(years=0), ss.datedur(years=10), ss.datedur(years=1)) + ss.date(2000)
    # ss.dur.arange(ss.years(0), ss.years(10), ss.years(1)) + ss.datedur(years=1)
    # ss.dur.arange(ss.years(0), ss.years(10), ss.datedur(years=1)) + ss.datedur(years=1)
    # ss.dur.arange(ss.years(0), ss.datedur(years=10), ss.datedur(years=1)) + ss.datedur(years=1)
    ss.dur.arange(ss.years(0), ss.years(10), ss.years(1)) + ss.date(2000)
    ss.dur.arange(ss.years(0), ss.years(10), ss.datedur(years=1)) + ss.date(2000)
    ss.dur.arange(ss.years(0), ss.datedur(years=10), ss.datedur(years=1)) + ss.date(2000)

    # Rates
    assert (1/ss.years(1)) == ss.freqperyear(1)
    assert (2/ss.years(1)) == ss.freqperyear(2)
    assert (4/ss.years(1)) == ss.freqperyear(4)
    assert (4/ss.datedur(1)) == ss.freqperyear(4)
    assert (ss.freqperday(5)*ss.datedur(days=1)) == 5
    assert 2/ss.freqperyear(0.25) == ss.years(8)
    assert 1/(2*ss.freqperyear(0.25)) == ss.years(2)
    assert ss.freqperyear(0.5)/ss.freqperyear(1) == 0.5

    # Probabilities
    p = ss.prob(0.1, ss.datedur(years=1))
    f = lambda factor: 1 - np.exp(-(-np.log(1 - p.value))/factor)
    assert p*ss.datedur(years=2) == f(0.5)
    assert p * ss.years(0.5) == f(2)
    assert p * ss.datedur(months=1) == f(12)

    p = ss.prob(0.1, ss.years(1))
    assert p*ss.datedur(years=2) == f(0.5)
    assert p * ss.years(0.5 ) == f(2)
    assert p * ss.datedur(months=1) == f(12)

    p = ss.per(0.1, ss.datedur(years=1))
    f = lambda factor: 1 - np.exp(-p.value/factor)
    assert p*ss.datedur(years=2) == f(0.5)
    assert p * ss.years(0.5) == f(2)
    assert p * ss.datedur(months=1) == f(12)

    p = ss.per(0.1, ss.years(1))
    assert p*ss.datedur(years=2) == f(0.5)
    assert p * ss.years(0.5) == f(2)
    assert p * ss.datedur(months=1) == f(12)

    return p


@sc.timer()
def test_datearray_operations():
    sc.heading('Test DateArray add/sub operations')

    a = ss.DateArray([ss.date('2020-01-01'), ss.date('2021-01-01')])
    b = ss.DateArray([ss.years(1), ss.years(2)])
    c = ss.years([1, 2])

    expected_dates_plus_month = ss.DateArray([ss.date('2020-01-31'), ss.date('2021-01-31')])
    assert np.array_equal(a + ss.months(1), expected_dates_plus_month)
    assert np.array_equal(ss.months(1) + a, expected_dates_plus_month)

    expected_months_plus = ss.DateArray([ss.months(13), ss.months(25)])
    assert np.array_equal(ss.months(1) + b, expected_months_plus)
    assert np.array_equal(b + ss.months(1), expected_months_plus)

    expected_months_plus_float = ss.months([13, 25])
    assert np.array_equal(ss.months(1) + c, expected_months_plus_float)
    expected_years_plus = ss.years([1 + 1/12, 2 + 1/12])
    assert np.array_equal(c + ss.months(1), expected_years_plus)

    with pytest.raises(TypeError):
        _ = a + ss.date('2020-01-01')
    with pytest.raises(TypeError):
        _ = ss.date('2020-01-01') + a

    expected_dates_plus = ss.DateArray([ss.date('2021-01-01'), ss.date('2022-01-01')])
    assert np.array_equal(b + ss.date('2020-01-01'), expected_dates_plus)
    assert np.array_equal(ss.date('2020-01-01') + b, expected_dates_plus)
    assert np.array_equal(c + ss.date('2020-01-01'), expected_dates_plus)
    assert np.array_equal(ss.date('2020-01-01') + c, expected_dates_plus)

    expected_dates_minus_month = ss.DateArray([ss.date('2019-12-02'), ss.date('2020-12-02')])
    assert np.array_equal(a - ss.months(1), expected_dates_minus_month)

    with pytest.raises(TypeError):
        _ = ss.months(1) - a

    expected_months_minus = ss.DateArray([ss.months(-11), ss.months(-23)])
    assert np.array_equal(ss.months(1) - b, expected_months_minus)

    expected_months_sub = ss.DateArray([ss.months(11), ss.months(23)])
    assert np.array_equal(b - ss.months(1), expected_months_sub)
    expected_months_minus_float = ss.months([-11, -23])
    assert np.array_equal(ss.months(1) - c, expected_months_minus_float)
    expected_years_minus = ss.years([1 - 1/12, 2 - 1/12])
    assert np.array_equal(c - ss.months(1), expected_years_minus)

    expected_datedur_from_date = ss.DateArray([ss.datedur(years=0), ss.datedur(years=-1)])
    assert np.array_equal(ss.date('2020-01-01') - a, expected_datedur_from_date)

    expected_datedur_from_array = ss.DateArray([ss.datedur(years=0), ss.datedur(years=1)])
    assert np.array_equal(a - ss.date('2020-01-01'), expected_datedur_from_array)

    with pytest.raises(TypeError):
        _ = b - ss.date('2020-01-01')
    with pytest.raises(TypeError):
        _ = c - ss.date('2020-01-01')

    expected_dates_from_date = ss.DateArray([ss.date('2019-01-01'), ss.date('2018-01-01')])
    assert np.array_equal(ss.date('2020-01-01') - b, expected_dates_from_date)
    assert np.array_equal(ss.date('2020-01-01') - c, expected_dates_from_date)

    return


# Run as a script
if __name__ == '__main__':
    T = sc.timer('\nTotal time')

    o1 = test_ratio()
    o2 = test_classes()
    o3 = test_callable_dists()
    o4 = test_syntax()
    o5 = test_datearray_operations()

    T.toc()



"""
Example SIR model in LASER, from laser_core/docs/example.rst

Requires laser_core~=0.5.1
"""
import numpy as np
import matplotlib.pyplot as plt
from laser_core import LaserFrame
from laser_core import PropertySet
import sciris as sc

class SIRModel:
    def __init__(self, params):
        # Model Parameters
        self.params = params

        # Initialize the population LaserFrame
        self.population = LaserFrame(capacity=params.population_size,initial_count=params.population_size)

        # Add disease state property (0 = Susceptible, 1 = Infected, 2 = Recovered)
        self.population.add_scalar_property("disease_state", dtype=np.int32, default=0)

        # Add a recovery timer property (for intrahost progression, optional for timing)
        self.population.add_scalar_property("recovery_timer", dtype=np.int32, default=0)

        # Results tracking
        self.results = LaserFrame( capacity = 1 ) # number of nodes
        self.results.add_vector_property( "S", length=params["timesteps"], dtype=np.float32 )
        self.results.add_vector_property( "I", length=params["timesteps"], dtype=np.float32 )
        self.results.add_vector_property( "R", length=params["timesteps"], dtype=np.float32 )

        # Components
        self.components = []

    def add_component(self, component):
        self.components.append(component)

    def track_results(self, tick):
        susceptible = (self.population.disease_state == 0).sum()
        infected = (self.population.disease_state == 1).sum()
        recovered = (self.population.disease_state == 2).sum()
        total = self.population.count
        self.results.S[tick] = susceptible / total
        self.results.I[tick] = infected / total
        self.results.R[tick] = recovered / total

    def run(self):
        for tick in range(self.params.timesteps):
            for component in self.components:
                component.step()
            self.track_results(tick)

    def plot_results(self):
        plt.figure(figsize=(10, 6))
        plt.plot(self.results.S, label="Susceptible (S)", color="blue")
        plt.plot(self.results.I, label="Infected (I)", color="red")
        plt.plot(self.results.R, label="Recovered (R)", color="green")
        plt.title("SIR Model Dynamics with LASER Components")
        plt.xlabel("Time (Timesteps)")
        plt.ylabel("Fraction of Population")
        plt.legend()
        plt.grid()
        plt.show()
        plt.savefig("gpt_sir.png")


class IntrahostProgression:
    def __init__(self, model):
        self.population = model.population

        # Seed the infection
        num_initial_infected = int(0.01 * model.params.population_size)  # e.g., 1% initially infected
        infected_indices = np.random.choice(model.params.population_size, size=num_initial_infected, replace=False)
        self.population.disease_state[infected_indices] = 1

        # Initialize recovery timer for initially infected individuals
        initially_infected = self.population.disease_state == 1
        self.population.recovery_timer[initially_infected] = np.random.randint(5, 15, size=initially_infected.sum())

    def step(self):
        infected = self.population.disease_state == 1

        # Decrement recovery timer
        self.population.recovery_timer[infected] -= 1

        # Recover individuals whose recovery_timer has reached 0
        recoveries = infected & (self.population.recovery_timer <= 0)
        self.population.disease_state[recoveries] = 2


class Transmission:
    def __init__(self, model):
        self.population = model.population
        self.infection_rate = model.params.infection_rate

    def step(self):
        susceptible = self.population.disease_state == 0
        infected = self.population.disease_state == 1

        num_susceptible = susceptible.sum()
        num_infected = infected.sum()
        population_size = len(self.population)

        # Fraction of infected and susceptible individuals
        fraction_infected = num_infected / population_size

        # Transmission logic: Probability of infection per susceptible individual
        infection_probability = self.infection_rate * fraction_infected

        # Apply infection probability to all susceptible individuals
        new_infections = np.random.rand(num_susceptible) < infection_probability

        # Set new infections and initialize their recovery_timer
        susceptible_indices = np.where(susceptible)[0]
        newly_infected_indices = susceptible_indices[new_infections]
        self.population.disease_state[newly_infected_indices] = 1
        self.population.recovery_timer[newly_infected_indices] = np.random.randint(5, 15, size=newly_infected_indices.size)  # Random recovery time


def run_sim(run=True, plot=False):
    """ Create and run the simulation """
    params = PropertySet({
        "population_size": 100_000*2,
        "infection_rate": 0.3,
        "timesteps": 100*10
    })

    # Initialize the model
    sir_model = SIRModel(params)

    # Initialize and add components
    sir_model.add_component(IntrahostProgression(sir_model))
    sir_model.add_component(Transmission(sir_model))

    # Run the simulation
    if run:
        sir_model.run()

    # Plot results
    if plot:
        sir_model.plot_results()
    return sir_model


if __name__ == '__main__':
    with sc.timer():
        sir_model = run_sim()
    sir_model.plot_results()


"""
Example maternal infection with congenital outcomes.

Demonstrates the generic congenital outcome framework in the base ``Infection``
class. ``CongenitalDisease`` is a simple SIR that assigns birth outcomes
(stillborn, congenital infection, normal) to unborn agents infected via
``PrenatalNet``.

How the congenital framework works:
    1. Mother-to-child transmission happens via ``PrenatalNet`` — the base
       ``Infection.infect()`` calls ``set_congenital()`` when the target is
       an unborn agent (age < 0).
    2. ``set_congenital()`` samples an outcome from ``birth_outcomes`` and
       schedules it at delivery time by setting ``ti_<outcome>``.
    3. ``step_congenital()`` (called each timestep from ``step_state()``)
       checks whether any scheduled outcomes are due. Death outcomes
       ('stillborn', 'nnd', 'miscarriage') call ``request_death()``;
       non-lethal outcomes set a BoolArr state (e.g. ``congenital = True``).

To use the framework in your own disease:
    1. Define ``birth_outcome_keys`` and ``birth_outcomes`` in pars
    2. Define matching ``ti_<key>`` FloatArr states for each outcome
    3. Define BoolArr states for non-lethal outcomes (same name as the key)
    4. Optionally define ``cs_outcome`` FloatArr to store outcome indices
    5. Call ``self.step_congenital()`` from ``step_state()``

For diseases with state- or gestational-age-dependent outcome probabilities
(e.g. syphilis, where outcomes differ by infection stage), override
``_assign_congenital_outcomes()`` and provide multiple keyed distributions
in ``birth_outcomes``.

Usage::

    import starsim as ss
    import starsim_examples as sse

    sim = ss.Sim(
        diseases=sse.CongenitalDisease(beta=0.2, init_prev=0.2),
        demographics=[ss.Pregnancy(fertility_rate=ss.freqperyear(30), burnin=True), ss.Deaths()],
        networks=[ss.PrenatalNet(), ss.RandomNet()],
    )
    sim.run()

    # Check results
    print('Stillbirths:', sim.results.pregnancy.stillbirths.sum())
    print('Congenital infections:', sim.diseases.congenitaldisease.congenital.sum())
"""

import numpy as np
import sciris as sc
import starsim as ss

__all__ = ['CongenitalDisease']


class CongenitalDisease(ss.SIR):
    """
    Simple disease with congenital outcomes via the generic framework.

    Infected mothers transmit to their unborn via PrenatalNet. At transmission,
    the base ``set_congenital()`` samples an outcome (stillborn, congenital,
    normal) and schedules it for delivery time. ``step_congenital()``
    in ``step_state()`` executes those scheduled events.

    Pars:
        birth_outcome_keys (list): outcome names — each needs a ``ti_<key>`` state
        birth_outcomes (objdict):  ``ss.choice`` distributions keyed by category;
            use 'default' for a single distribution applied to all infections

    States:
        ti_stillborn (FloatArr):  timestep when stillbirth fires (triggers request_death)
        ti_congenital (FloatArr): timestep when congenital infection fires (sets BoolArr)
        ti_normal (FloatArr):     timestep when normal outcome fires (no effect)
        congenital (BoolArr):     True if agent has congenital infection
        cs_outcome (FloatArr):    index into birth_outcome_keys for each agent
    """

    def __init__(self, pars=None, **kwargs):
        super().__init__()
        self.define_pars(
            # Keys define the possible outcomes; probabilities sum to 1
            birth_outcome_keys = ['stillborn', 'congenital', 'normal'],
            birth_outcomes     = sc.objdict(
                default=ss.choice(a=3, p=np.array([0.3, 0.4, 0.3])),
            ),
        )
        self.update_pars(pars, **kwargs)
        self.define_states(
            ss.FloatArr('ti_stillborn'),   # Scheduled timestep for stillbirth
            ss.FloatArr('ti_congenital'),  # Scheduled timestep for congenital infection
            ss.FloatArr('ti_normal'),      # Scheduled timestep for normal outcome
            ss.BoolArr('congenital'),      # Whether agent has congenital infection
            ss.FloatArr('cs_outcome'),     # Outcome index (0=stillborn, 1=congenital, 2=normal)
        )
        return

    def step_state(self):
        super().step_state()
        self.step_congenital()  # Process any outcomes scheduled for this timestep
        return



"""
Fetal health connector and treatment intervention examples.

These modules demonstrate how to connect a disease to fetal health outcomes
and how to implement treatment that reverses fetal damage. They are designed
to be extended for real applications.

Architecture:
    - ``fetal_infection`` (Connector): watches for infections in pregnant women
      and applies fetal damage (preterm risk via timing shifts, low birth weight
      via growth restriction). Damage is applied both at conception (if already
      infected) and during pregnancy (if newly infected).
    - ``treat_pregnant`` (Intervention): treats infected pregnant women within a
      specified year range and partially reverses fetal damage.

Both modules require ``ss.FetalHealth()`` in the sim's ``custom`` modules and
``ss.Pregnancy()`` in demographics. The connector also requires the target
disease (default: SIR) to be present.

Usage::

    import starsim as ss
    import starsim_examples as sse

    sim = ss.Sim(
        diseases=ss.SIR(beta=0.1),
        demographics=[ss.Pregnancy(fertility_rate=ss.freqperyear(30)), ss.Deaths()],
        connectors=sse.fetal_infection(),
        interventions=sse.treat_pregnant(disease='sir', start_year=2025),
        custom=ss.FetalHealth(),
        networks=[ss.PrenatalNet(), ss.RandomNet()],
    )
    sim.run()

To extend for a different disease, subclass ``fetal_infection`` and override
``_apply_damage`` with disease-specific logic (e.g. stage-dependent penalties).
To extend ``treat_pregnant``, subclass and override ``step()`` with custom
eligibility criteria or reversal logic.
"""

import starsim as ss

__all__ = ['fetal_infection', 'treat_pregnant']


class treat_pregnant(ss.Intervention):
    """
    Treat infected pregnant women and partially reverse fetal damage.

    Each timestep, identifies pregnant women infected with the target disease,
    treats a fraction of them (curing infection), and reverses a portion of
    the accumulated fetal damage (growth restriction and timing shift).

    Treatment only applies between ``start_year`` and ``end_year``. If not
    specified, defaults to the full simulation period.

    Args:
        disease (str):          name of the target disease (default: 'sir')
        start_year (float):     first year treatment is available (default: sim start)
        end_year (float):       last year treatment is available (default: sim end)

    Pars:
        p_treat (Dist):             probability of treating an eligible woman per timestep
        tx_growth_reversal (float): fraction of growth restriction to reverse (0-1)
        tx_timing_reversal (float): fraction of timing shift to reverse (0-1)

    States:
        ti_treated (FloatArr):  timestep when each agent was treated

    Example — treatment starting in 2025 with 50% coverage::

        sse.treat_pregnant(disease='sir', start_year=2025, p_treat=ss.bernoulli(p=0.5))
    """

    def __init__(self, disease='sir', start_year=None, end_year=None, **kwargs):
        super().__init__(**kwargs)
        self.disease_name = disease
        self.start_year = start_year
        self.end_year   = end_year
        self.define_pars(
            p_treat            = ss.bernoulli(p=0.9),
            tx_growth_reversal = 0.7,
            tx_timing_reversal = 0.7,
        )
        self.define_states(
            ss.FloatArr('ti_treated', label='Time of treatment'),
        )
        return

    def init_pre(self, sim):
        super().init_pre(sim)

        # Validate that the required disease and FetalHealth modules exist
        if self.disease_name not in sim.diseases:
            raise ValueError(f'treat_pregnant requires disease "{self.disease_name}" but it was not found in the sim.')
        if 'fetal_health' not in sim.custom:
            raise ValueError('treat_pregnant requires a FetalHealth module. Add ss.FetalHealth() to custom.')

        # Default year bounds to the full sim period
        if self.start_year is None: self.start_year = sim.t.start
        if self.end_year is None:   self.end_year = sim.t.stop
        return

    def step(self):
        # Only apply treatment within the active year range
        year = self.t.now('year')
        if year < self.start_year or year > self.end_year:
            return

        sim     = self.sim
        preg    = sim.people.pregnancy
        disease = sim.diseases[self.disease_name]
        fh      = sim.custom['fetal_health']

        # Find pregnant women infected with the target disease
        eligible = preg.pregnant & disease.infected
        treated = self.pars.p_treat.filter(eligible)
        if len(treated):
            # Cure infection
            disease.infected[treated]  = False
            disease.recovered[treated] = True
            self.ti_treated[treated]   = self.ti

            # Partially reverse fetal damage from the infection
            fh.reverse_growth_restriction(treated, self.pars.tx_growth_reversal)
            fh.reverse_timing_shift(treated, self.pars.tx_timing_reversal)
        return


class fetal_infection(ss.Connector):
    """
    Connect a disease to fetal health outcomes during pregnancy.

    Monitors for infections in pregnant women and applies fetal damage:
    - **Timing shift**: brings delivery forward (increases preterm birth risk).
      Drawn from a lognormal distribution per affected pregnancy.
    - **Growth restriction**: reduces birth weight by a fixed fractional penalty.

    Damage is applied at two points:
    1. At conception, if the mother is already infected (via a conception callback
       registered with FetalHealth).
    2. During pregnancy, when a new infection occurs (detected in ``step()``
       by checking ``ti_infected == self.ti``).

    Requires ``ss.FetalHealth()`` in ``custom`` and an SIR disease in ``diseases``.

    Pars:
        timing_shift (Dist):    weeks to shift delivery forward per infection (default: lognormal mean=3, std=1)
        growth_penalty (float): fractional birth weight reduction per infection (default: 0.15 = 15%)

    To adapt for a different disease::

        class my_fetal_connector(fetal_infection):
            def init_pre(self, sim):
                # Register with FetalHealth, but check for your disease instead
                ...

            def _apply_damage(self, uids):
                # Custom damage logic, e.g. stage-dependent penalties
                fh = self.sim.custom['fetal_health']
                disease = self.sim.diseases.my_disease
                severe = disease.severe[uids]
                mild_uids   = uids[~severe]
                severe_uids = uids[severe]
                fh.apply_growth_restriction(mild_uids, 0.05)
                fh.apply_growth_restriction(severe_uids, 0.25)
                ...
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.define_pars(
            timing_shift   = ss.lognorm_ex(mean=3.0, std=1.0),
            growth_penalty = 0.15,
        )
        return

    def init_pre(self, sim):
        super().init_pre(sim)

        # Validate that the required modules exist
        if 'fetal_health' not in sim.custom:
            raise ValueError('fetal_infection requires a FetalHealth module. Add ss.FetalHealth() to custom.')
        if 'sir' not in sim.diseases:
            raise ValueError('fetal_infection requires an SIR disease module.')

        # Register a callback so we can apply damage at conception for
        # women who are already infected when they become pregnant
        fh = sim.custom['fetal_health']
        fh.add_conception_callback(self._on_conception)
        return

    def _on_conception(self, uids):
        """ Called by FetalHealth when new pregnancies begin — damage if already infected """
        infected = self.sim.diseases.sir.infected[uids]
        infected_uids = uids[infected]
        if len(infected_uids):
            self._apply_damage(infected_uids)
        return

    def _apply_damage(self, uids):
        """
        Apply fetal damage to pregnancies of infected women.

        Override this method to customize the damage logic for a different
        disease (e.g. stage-dependent growth penalties, trimester-dependent
        timing shifts).
        """
        fh = self.sim.custom['fetal_health']
        shifts = self.pars.timing_shift.rvs(uids)
        fh.apply_timing_shift(uids, shifts)
        fh.apply_growth_restriction(uids, self.pars.growth_penalty)
        return

    def step(self):
        """ Each timestep, check for new infections in pregnant women and apply damage """
        sim  = self.sim
        preg = sim.people.pregnancy
        if not preg.pregnant.any():
            return

        # Find pregnant women newly infected this timestep
        pregnant_uids  = preg.pregnant.uids
        newly_infected = sim.diseases.sir.ti_infected == self.ti
        affected = pregnant_uids[newly_infected[pregnant_uids]]
        if len(affected):
            self._apply_damage(affected)
        return



"""
Utilities to help with debugging Starsim runs
"""
import sys
import types
import platform
import numpy as np
import numba as nb
import pandas as pd
import sciris as sc
import matplotlib as mpl
import starsim as ss

__all__ = ['Profile', 'Debugger', 'Diagnostics', 'check_requires', 'check_version',
           'metadata', 'mock_time', 'mock_sim', 'mock_people', 'mock_module']


class Profile(sc.profile):
    """
    Class to profile the performance of a simulation

    Typically invoked via `sim.profile()`.

    Args:
        sim (`ss.Sim`): the sim to profile
        follow (func/list): a list of functions/methods to follow in detail
        do_run (bool): whether to immediately run the sim
        plot (bool): whether to plot time spent per module step
        **kwargs (dict): passed to `sc.profile()`

    **Example**:

        import starsim as ss

        net = ss.RandomNet()
        sis = ss.SIS()
        sim = ss.Sim(networks=net, diseases=sis)
        prof = sim.profile(follow=[net.add_pairs, sis.infect])
        prof.disp()
    """
    def __init__(self, sim, follow=None, do_run=True, plot=True, verbose=True, **kwargs):
        assert isinstance(sim, ss.Sim), f'Only an ss.Sim object can be profiled, not {type(sim)}'
        if 'skipzero' not in kwargs:
            kwargs['skipzero'] = True # Since provide the same follow to both sim.init() and sim.run(), so expect zero entries
        super().__init__(run=None, do_run=False, verbose=verbose, **kwargs)
        self.orig_sim = sim
        self.ss_follow = follow # Keep a copy to know if we should overwrite this

        # Initialize: copy the sim and time initialization
        sim = self.orig_sim.copy() # Copy so the sim can be reused
        self.sim = sim
        self.run_func = sim.run # This is used internally by sc.profile.run()
        self.init_prof = None # Profiling the initialization

        # Optionally run
        if do_run:
            self.profile_init()
            self.run()
            if plot:
                self.plot_cpu()
        return

    def profile_init(self):
        """ Handle sim init -- both run it and profile it """
        if not self.sim.initialized:
            self.init_prof = sc.profile(self.sim.init, follow=self.ss_follow, verbose=False, skipzero=True)
        else:
            errormsg = 'Cannot profile initialization of already initialized sim'
            raise RuntimeError(errormsg)
        return self.init_prof

    def run(self, disp=None):
        """ Profile the performance of the simulation """
        sim = self.sim
        if not self.sim.initialized:
            self.sim.init() # Don't profile

        # Get the functions from the initialized sim
        if self.ss_follow is None:
            loop_funcs = [e['func'] for e in sim.loop.funcs]
            self.follow = [sim.run] + loop_funcs
        else:
            self.follow = self.ss_follow

        # Run the profiling on the sim run
        super().run(disp=disp)

        # Add initialization to the other timings
        if self.init_prof:
            self.merge(self.init_prof, inplace=True, swap=True)

        return self

    def disp(self, bytime=1, maxentries=10, skiprun=True):
        """ Same as sc.profile.disp(), but skip the run function by default """
        return super().disp(bytime=bytime, maxentries=maxentries, skiprun=skiprun)

    def plot_cpu(self):
        """ Shortcut to sim.loop.plot_cpu() """
        self.sim.loop.plot_cpu()
        return


class Debugger(sc.prettyobj):
    """
    Step through one or more sims and pause or raise an exception when a condition is met

    Args:
        args (list): the sim or sims to step through
        func (func/str): the function to run on the sims (or can use a built-in one, e.g. 'equal')
        skip (list): if provided, additional object names/types to skip (not check)
        verbose (bool): whether to print progress during the run
        die (bool): whether to raise an exception if the condition is met; alternatively 'pause' will pause execution, and False will just print
        run (bool): whether to run immediately

    **Examples**:

        ## Example 1: Identical sims are identical
        import starsim as ss

        s1 = ss.Sim(pars=dict(diseases='sis', networks='random'), n_agents=250)
        s2 = s1.copy()
        s3 = s1.copy()
        db = ss.Debugger(s1, s2, s3, func='equal')
        db.run()

        ## Example 2: Pause when sim results diverge
        import sciris as sc
        import starsim as ss

        # Set up the sims
        pars = dict(networks='random', n_agents=1000, verbose=0)
        sis1 = ss.SIS(beta=0.050)
        sis2 = ss.SIS(beta=0.051) # Very slightly more
        s1 = ss.Sim(pars, diseases=sis1)
        s2 = ss.Sim(pars, diseases=sis2)

        # Run the debugger
        db = ss.Debugger(s1, s2, func='equal_results', die='pause')
        db.run()

        # Show non-matching results
        sc.heading('Differing results')
        df = db.results[-1].df
        df = df[~df['equal']]
        df.disp()
    """
    def __init__(self, *args, func, skip=None, verbose=True, die=True, run=False):
        default_skip = [
            ss.Sim, ss.Module, ss.Dist, '_states',
            types.FunctionType, types.MethodType,
            types.BuiltinFunctionType, types.BuiltinMethodType,
            types.WrapperDescriptorType, types.MethodDescriptorType,
            staticmethod, classmethod
        ]
        self.sims = args
        for sim in self.sims:
            if not sim.initialized:
                sim.init()
        self.verbose = verbose
        self.die = die
        self.process_func(func)
        self.until = self.sims[0].t.npts
        self.ti = self.sims[0].ti
        self.results = sc.objdict()
        self.skip = sc.mergelists(skip, default_skip)
        self.kw = dict(skip=self.skip, detailed=True, leaf=True)
        self.pause = False
        if run:
            self.run()
        return

    def process_func(self, func):
        if callable(func):
            self.func = func
        elif isinstance(func, str):
            if hasattr(self, func):
                self.func = getattr(self, func)
            else:
                errormsg = f'Unrecognized function "{func}"'
                raise ValueError(errormsg)
        else:
            errormsg = f'Unrecognized function type "{type(func)}"'
            raise ValueError(errormsg)

    def equal_check(self, e, which):
        """ Check if equality is false, and print a message or die """
        if not e.eq:
            msg = f'Found a difference in {which} on ti={self.ti}:\n'
            msg += f'{e.df}'
            if self.die == True:
                raise RuntimeError(msg)
            elif self.die == 'pause':
                self.pause = True
            print(msg)
        else:
            msg = f'No differences in {which}'
        return msg

    def equal_dists(self, *sims):
        """ Check if the dists are equal """
        dstates = []
        for sim in sims:
            ds = {d.trace:d.show_state(output=True) for d in sim.dists.dists.values()}
            dstates.append(ds)
        e = sc.Equal(*dstates, **self.kw)
        self.equal_check(e, 'dists')
        return e

    def equal_pars(self, *sims, skip='label'):
        """ Check if SimPars are equal """
        skip = sc.tolist(skip)
        parslist = [s.pars.copy() for s in sims]
        for key in skip: # Do not need to do on every step, but so fast it doesn't matter
            for pars in parslist:
                pars.pop(key, None)

        e = sc.Equal(*parslist, **self.kw)
        self.equal_check(e, 'pars')
        return e

    def equal_people(self, *sims):
        """ Check if people are equal """
        e = sc.Equal(*[s.people.states for s in sims], **self.kw)
        self.equal_check(e, 'people')
        return e

    def equal_networks(self, *sims):
        ncs = [[n.edges for n in s.networks.values()] for s in sims]
        e = sc.Equal(*ncs, **self.kw)
        self.equal_check(e, 'network edges')
        return e

    def equal_results(self, *sims):
        e = sc.Equal(*[s.results for s in sims], **self.kw)
        self.equal_check(e, 'results')
        return e

    def equal(self, *sims):
        """ Run all other tests """
        out = sc.objdict()
        out.dists    = self.equal_dists(*sims)
        out.pars     = self.equal_pars(*sims)
        out.people   = self.equal_people(*sims)
        out.networks = self.equal_networks(*sims)
        out.results  = self.equal_results(*sims)
        if self.verbose:
            for k,e in out.items():
                sc.printgreen('—'*80)
                sc.printgreen(f'{k}:')
                e.df.disp()
                print()
        return out

    def check(self):
        self.results[f'{self.ti}'] = self.func(*self.sims)
        return

    def step(self, reseed=True):
        self.ti += 1
        if self.verbose: sc.heading(f'Working on step ti={self.ti}')
        for sim in self.sims:
            if reseed:
                np.random.seed(sim.pars.rand_seed + self.ti) # Reset the random seed, in case anything is not CRN-safe
            sim.run_one_step()
        self.check()
        return

    def run(self, reseed=True):
        self.pause = False # Reset so we can run
        self.check()
        while self.ti < self.until:
            self.step(reseed=reseed)
            if self.pause: # If we hit a checkpoint, pause and break the loop
                break
        if not self.pause:
            for sim in self.sims:
                sim.finalize()
            if len(self.sims) > 1:
                self.df = ss.diff_sims(self.sims[0], self.sims[1], full=True, output=True)
        return


class Diagnostics(sc.quickobj):
    """ Helper class for storing detailed diagnostic information; see sim.set_diagnostics() for usage; not to be used directly

    See also ss.Debugger() for a user API for debugging multiple sims.
    """
    def __init__(self, sim, rvs=True, states=True, detailed=False):
        self.sim = sim
        self.rvs = None
        self.states = None
        self.detailed = detailed
        if rvs:
            self.rvs = sc.objdict()
            self.rvs_file = rvs if isinstance(rvs, str) else 'debug_rvs.json'
        if states:
            self.states = sc.objdict()
            self.states_file = states if isinstance(states, str) else 'debug_states.json'
        return

    @staticmethod
    def compute_stats(v):
        """ Compute length, mean, min, max, and unique hash """
        stats = sc.objdict(
            n = len(v),
            mean = v.mean(),
            min = v.min(),
            max = v.max(),
            hash = sc.sha(v).hexdigest()[:16],
        )
        return stats

    @property
    def ti_key(self):
        return f'ti{self.sim.ti}'

    def store_rvs(self, dist, rvs):
        """ Store random numbers (random variates) """
        entry = self.rvs.setdefault(self.ti_key, sc.objdict())

        if self.detailed:
            entry[dist.trace] = rvs
        else:
            entry[dist.trace] = self.compute_stats(rvs)
        return

    def store_states(self, key=None):
        """ Store all states from people """
        key = sc.ifelse(key, self.ti_key)
        entry = sc.objdict()
        for key, state in self.sim.people.states.items():
            if self.detailed:
                entry[key] = state.values
            else:
                entry[key] = self.compute_stats(state)

        self.states[key] = entry
        return

    def export(self):
        """ Export saved data to file(s); filenames must be set when the object is created """
        if self.rvs:
            sc.savejson(filename=self.rvs_file, obj=self.rvs)
        if self.states:
            sc.savejson(filename=self.states_file, obj=self.states)
        return


def check_requires(sim, requires, *args):
    """ Check that the module's requirements (of other modules) are met """
    errs = sc.autolist()
    all_classes = [m.__class__ for m in sim.modules]
    all_names = [m.name for m in sim.modules]
    for req in sc.mergelists(requires, *args):
        if req not in all_classes + all_names:
            errs += req
    if len(errs):
        errormsg = f'The following module(s) are required, but the Sim does not contain them: {sc.strjoin(errs)}'
        raise AttributeError(errormsg)
    return


def check_version(expected, die=False, warn=True):
    """
    Check the expected Starsim version with the one actually installed. The expected
    version string may optionally start with '>=' or '<=' (== is implied otherwise),
    but other operators (e.g. ~=) are not supported. Note that '>' and '<' are interpreted
    to mean '>=' and '<='; '>' and '<' are not supported.

    Args:
        expected (str): expected version information
        die (bool): whether or not to raise an exception if the check fails
        warn (bool): whether to raise a warning if the check fails

    **Example**:

        ss.check_version('>=3.0.0', die=True) # Will raise an exception if an older version is used
    """
    if   expected.startswith('>'): valid = [0,1]
    elif expected.startswith('<'): valid = [0,-1]
    elif expected.startswith('!'): valid = [1,-1]
    else: valid = [0] # Assume == is the only valid comparison
    expected = expected.lstrip('<=>') # Remove comparator information
    version = ss.__version__
    compare = sc.compareversions(version, expected) # Returns -1, 0, or 1
    relation = ['older', '', 'newer'][compare+1] # Picks the right string
    if relation: # Versions mismatch, print warning or raise error
        string = f'Starsim is {relation} than expected ({version} vs. {expected})'
        if compare not in valid:
            if die:
                raise ValueError(string)
            elif warn:
                ss.warn(string)
    return compare


def metadata(comments=None):
    """ Store metadata; like `sc.metadata()`, but optimized for speed """
    md = sc.objdict(
        version = ss.__version__,
        versiondate = ss.__versiondate__,
        timestamp = sc.getdate(),
        user      = sc.getuser(),
        system = sc.objdict(
            platform   = platform.platform(),
            executable = sys.executable,
            version    = sys.version,
        ),
        versions = sc.objdict(
            python     = platform.python_version(),
            numpy      = np.__version__,
            numba      = nb.__version__,
            pandas     = pd.__version__,
            sciris     = sc.__version__,
            matplotlib = mpl.__version__,
            starsim    = ss.__version__,
        ),
        comments = comments,
    )
    return md


def mock_time(dt=1.0, dur=10, start=2000):
    """ Create a minimal mock "Time" object """
    t = sc.objdict(
        dt = dt,
        ti = 0,
        start = start,
        stop = None,
        dur = dur,
        is_absolute = True,
        initialized = True,
    )
    return t


def mock_sim(n_agents=100, **kwargs):
    """
    Create a minimal mock "Sim" object to initialize objects that require it

    Args:
        n_agents (int): the number of agents to create
        **kwargs (dict): passed to `ss.mock_time()`
    """
    t = mock_time(**kwargs)
    sim = sc.objdict(
        label = 'mock_sim',
        people = mock_people(n_agents),
        t = t,
        dt = kwargs.get('dt', ss.years(1.0)),
        pars = t,
        results = sc.objdict(),
        networks = sc.objdict(), # Needed for infections
        analyzers = sc.objdict(), # Needed for infection log
        diagnostics = None,
    )
    return sim


def mock_people(n_agents=100):
    """ Create a minimal mock "People" object """
    people = sc.objdict(
        uid = np.arange(n_agents),
        auids = np.arange(n_agents),
        slot = np.arange(n_agents),
        age = np.random.uniform(0, 70, size=n_agents),
        add_module = lambda x: None, # Placeholder function
    )
    return people


def mock_module(dur=10, **kwargs):
    """ Create a minimal mock "Module" object """
    mod = sc.objdict(
        name = 'mock_module',
        t = mock_time(**kwargs),
        dt = kwargs.get('dt', ss.years(1.0)),
    )
    return mod


"""
Test run, including ss.MultiSim and ss.parallel()
"""

# %% Imports and settings
import starsim as ss
import sciris as sc
import numpy as np
import matplotlib.pyplot as plt

n_agents = 1_000
do_plot = False
sc.options(interactive=False) # Assume not running interactively


def make_sim_pars(beta=0.1, **kwargs):
    pars = sc.objdict(
        n_agents = n_agents,
        demographics = True,
        networks = 'random',
        diseases = sc.objdict(type='sis', beta=beta), # To allow for modification later
        **kwargs,
    )
    return pars


@sc.timer()
def test_parallel():
    """ Test running two identical sims in parallel """
    sc.heading('Testing parallel...')
    pars = make_sim_pars()

    # Check that two identical sims match
    msim = ss.MultiSim([ss.Sim(pars, label='Sim1'), ss.Sim(pars, label='Sim2')])
    msim.run(shrink=False)
    s1, s2 = msim.sims
    assert np.allclose(s1.summary[:], s2.summary[:], rtol=0, atol=0, equal_nan=True), "Sims don't match and should"

    # Check that two non-identical sims don't match
    pars2 = sc.dcp(pars)
    pars2.diseases.beta *= 2
    sims = ss.MultiSim([ss.Sim(pars, label='Sim1'), ss.Sim(pars2, label='Sim2')])
    sims.run(shrink=False)
    s1, s2 = sims.sims
    assert not np.allclose(s1.summary[:], s2.summary[:], rtol=0, atol=0, equal_nan=True), "Sims do match and shouldn't"

    return s1, s2


@sc.timer()
def test_multisim():
    """ Check MultiSim methods """
    sc.heading('Testing MultiSim')
    pars = make_sim_pars()
    msim = ss.MultiSim(ss.Sim(pars))
    msim.init_sims()
    msim.run(n_runs=4)

    # Plot individual sims
    msim.plot()

    # Reduce and plot mean
    msim.mean()
    msim.plot()

    # Export results
    res_df = msim.results.to_df(resample='2YE')
    assert res_df.sis_n_susceptible_low.loc['2030-12-31'] == msim.results.sis_n_susceptible.low[29:31].mean()

    # Reduce and plot median
    msim.median()
    msim.plot()

    # Other methods
    msim.show()
    msim.summarize()
    msim.reset()

    return msim


@sc.timer()
def test_multisim_construction(do_plot=False):
    """ Test other MultiSim construction methods """
    sc.heading('Testing MultiSim construction')
    
    # Make sims
    def make_sims(n=2):
        sims = sc.autolist()
        for beta in 0.1*np.random.rand(n):
            pars = make_sim_pars(beta=beta, dur=20, rand_seed=np.random.randint(999))
            sims += ss.Sim(pars=pars, label=f'{beta=}')
        return sims
    
    # Construction options
    m1 = ss.MultiSim(make_sims(), label='list') # As list
    m2 = ss.MultiSim(tuple(make_sims()), label='tuple') # As tuple
    m3 = ss.MultiSim(m1.copy(), label='msim') # As MultiSim
    m4 = ss.MultiSim([m1.copy(), m2.copy()], label='multi-msim') # As list of MultiSims
    
    # Check that run works
    for msim in [m1, m2, m3, m4]:
        msim.run()
        if do_plot:
            msim.plot()
    
    return m4


@sc.timer()
def test_other():
    """ Check other run options """
    sc.heading('Testing other run options')
    # Check parallel
    pars = make_sim_pars()
    pars2 = sc.dcp(pars)
    pars2.diseases.beta *= 2
    s1 = ss.Sim(pars)
    s2 = ss.Sim(pars2)
    ss.parallel(s1, s2, inplace=False)
    assert not s1.initialized
    ss.parallel(s1, s2, inplace=True)
    assert not np.allclose(s1.summary[:], s2.summary[:], rtol=0, atol=0, equal_nan=True), "Sims do match and shouldn't"

    # Check single run
    s3 = ss.Sim(pars)
    s4 = ss.Sim(pars)
    s3 = ss.single_run(s3)
    s4 = ss.single_run(s4, n_agents=2000)
    assert not np.allclose(s3.summary[:], s4.summary[:], rtol=0, atol=0, equal_nan=True), "Sims do match and shouldn't"

    return s3,s4


if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)
    T = sc.timer()

    s1, s2 = test_parallel()
    msim = test_multisim()
    msim2 = test_multisim_construction(do_plot)
    s3,s4 = test_other()

    T.toc()

    if do_plot:
        plt.show()



#!/usr/bin/env python
"""
Run all the Jupyter notebook tutorials and user guide entries. Usage:
  ./check_notebooks.py

It will execute all notebooks in parallel, print out any errors
encountered, and print a summary of results.
"""
import sys
import docutils as ut
results = ut.execute_notebooks(*sys.argv[1:])



"""
Run tests of vaccines and products
"""
import sciris as sc
import numpy as np
import starsim as ss


def run_sir_vaccine(efficacy, leaky=True):
    sc.heading(f'Testing SIR vaccine with {efficacy = } and {leaky = }')

    # parameters
    v_frac      = 0.5    # fraction of population vaccinated
    total_cases = 500    # total cases at which point we check results
    tol         = 3      # tolerance in standard deviations for simulated checks

    # create a basic SIR sim
    sim = ss.Sim(
        n_agents = 1000,
        pars = dict(
          networks = dict(
                type = 'random',
                n_contacts = 4
          ),
          diseases = dict(
                type      = 'sir',
                init_prev = 0.01,
                dur_inf   = ss.years(10),
                p_death   = 0,
                beta      = ss.peryear(0.06),
          )
        ),
        dur = ss.years(10),
        dt  = ss.years(0.05)
    )
    sim.init(verbose=False)

    # work out who to vaccinate
    in_trial = sim.people.sir.susceptible.uids
    n_vac = round(len(in_trial) * v_frac)
    in_vac = np.random.choice(in_trial, n_vac, replace=False)
    in_pla = np.setdiff1d(in_trial, in_vac)
    uids = ss.uids(in_vac)

    # create and apply the vaccination
    vac = ss.simple_vx(efficacy=efficacy, leaky=leaky)
    vac.init_pre(sim)
    vac.administer(sim.people, uids)

    # check the relative susceptibility at the start of the simulation
    rel_susc = sim.people.sir.rel_sus.values
    assert min(rel_susc[in_pla]) == 1.0, 'Placebo arm is not fully susceptible'
    if not leaky:
        assert min(rel_susc[in_vac]) == 0.0, 'Nobody fully vaccinated (all_or_nothing)'
        assert max(rel_susc[in_vac]) == 1.0, 'Vaccine effective in everyone (all_or_nothing)'
        mean = n_vac * (1 - efficacy)
        sd = np.sqrt(n_vac * efficacy * (1 - efficacy))
        assert (np.mean(rel_susc[in_vac]) - mean) / sd < tol, 'Incorrect mean susceptibility in vaccinated (all_or_nothing)'
    else:
        assert max(abs(rel_susc[in_vac] - (1 - efficacy))) < 0.0001, 'Relative susceptibility not 1-efficacy (leaky)'

    # run the simulation until sufficient cases
    old_cases = []
    for idx in range(1000):
        sim.run_one_step()
        susc = sim.people.sir.susceptible.uids
        cases = np.setdiff1d(in_trial, susc)
        if len(cases) > total_cases:
            break
        old_cases = cases

    if len(cases) > total_cases:
        cases = np.concatenate([old_cases, np.random.choice(np.setdiff1d(cases, old_cases), total_cases - len(old_cases), replace=False)])
    vac_cases = np.intersect1d(cases, in_vac)

    # check to see whether the number of cases are as expected
    p = v_frac * (1 - efficacy) / (1 - efficacy * v_frac)
    mean = total_cases * p
    sd = np.sqrt(total_cases * p * (1 - p))
    assert (len(vac_cases) - mean) / sd < tol, 'Incorrect proportion of vaccincated infected'

    # for all or nothing check that fully vaccinated did not get infected
    if not leaky:
        assert len(np.intersect1d(vac_cases, in_vac[rel_susc[in_vac] == 1.0])) == len(vac_cases), 'Not all vaccine cases amongst vaccine failures (all or nothing)'
        assert len(np.intersect1d(vac_cases, in_vac[rel_susc[in_vac] == 0.0])) == 0, 'Vaccine cases amongst fully vaccincated (all or nothing)'

    return sim


@sc.timer()
def test_sir_vaccine_leaky():
    return run_sir_vaccine(0.3, False)


@sc.timer()
def test_sir_vaccine_all_or_nothing():
    return run_sir_vaccine(0.3, True)


@sc.timer()
def test_products(do_plot=False):
    sc.heading('Testing products')

    pars = sc.objdict(
        n_agents = 5e3,
        start = 2000,
        stop = 2020,
        diseases = 'sis',
        networks = 'random',
    )

    dx_data = sc.dataframe(
        columns =
            ['disease', 'state', 'result', 'probability'],
        data = [
            ['sis', 'susceptible', 'positive', 0.01],
            ['sis', 'susceptible', 'negative', 0.99],
            ['sis', 'infected', 'positive', 0.95],
            ['sis', 'infected', 'negative', 0.05],
        ]
    )

    # Using built-in products
    vx_start = 2005
    my_vaccine = ss.simple_vx(efficacy=0.9)
    vaccination = ss.routine_vx(
        product = my_vaccine,  # Product object
        prob = 0.8,
        start_year = vx_start,
    )

    # Using custom products
    dx_start = 2010
    screening = ss.routine_screening(
        product = ss.Dx(df=dx_data),
        prob = 0.9,
        start_year = dx_start,
    )

    # Run the sim
    sim = ss.Sim(pars, interventions=[screening, vaccination])
    sim.run()

    # Checks
    dxres = sim.results.routine_screening
    sisres = sim.results.sis
    y = sim.t.yearvec
    pre_dx = y < dx_start
    post_dx = y > dx_start
    pre_vx = y < vx_start
    post_vx = y > vx_start
    assert dxres.n_screened[pre_dx].sum() == 0, 'Expected no one screened before intervention start'
    assert dxres.n_screened[post_dx].sum() > 0, 'Expected people screened after intervention start'
    assert dxres.n_dx[pre_dx].sum() == 0, 'Expected no one diagnosed before intervention start'
    assert dxres.n_dx[post_dx].sum() > 0, 'Expected people diagnosed after intervention start'
    assert sisres.new_infections[pre_vx].mean() > sisres.new_infections[post_vx].mean(), 'Expected vaccine to reduce prevalence'

    if do_plot:
        sim.plot()

    return sim

class StatefulVx(ss.Vx):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.define_states(
            ss.BoolState('vaccinated'),  # True if a routine vaccine was ever delivered to this agent
        ),
        return

    def administer(self, people, uids):
        self.vaccinated[uids] = True
        self.results.n_vaccinated[self.ti] = len(uids)


@sc.timer()
def test_two_products():

    pars = sc.objdict(
        n_agents = 5e3,
        start = 2000,
        stop = 2020,
        diseases = 'sis',
        networks = 'random',
    )

    my_vaccine = StatefulVx()
    vx1 = ss.routine_vx(name='vx1', product = my_vaccine, prob=0.1,start_year = 2005, eligibility=lambda sim: (~sim.people.statefulvx.vaccinated).uids)
    vx2 = ss.routine_vx(name='vx2', product = my_vaccine, prob=0.2,start_year = 2005, eligibility=lambda sim: (~sim.people.statefulvx.vaccinated).uids)
    sim = ss.Sim(pars, interventions=[vx1,vx2])
    sim.run()

    assert (sim.people.vx1.vaccinated.sum()+sim.people.vx2.vaccinated.sum()) == sim.people.statefulvx.vaccinated.sum()

    return sim



if __name__ == '__main__':
    T = sc.timer()
    do_plot = True

    leaky  = test_sir_vaccine_leaky()
    a_or_n = test_sir_vaccine_all_or_nothing()
    prod   = test_products(do_plot=do_plot)
    prod2   = test_two_products()

    T.toc()



"""
Test Sim API
"""

# %% Imports and settings
import sys
import numpy as np
import sciris as sc
import matplotlib.pyplot as plt
import starsim as ss
import pytest

pars = sc.objdict(
    n_agents = 1_000
)
do_plot = False
sc.options(interactive=False) # Assume not running interactively


@sc.timer()
def test_single_defaults(do_plot=do_plot):
    """ Test a single MixingPool using defaults """
    sc.heading('Testing single mixing pool...')
    mp = ss.MixingPool()
    sir = ss.SIR()
    sim = ss.Sim(pars, diseases=sir, networks=mp)
    sim.run()
    res = sim.results.sir

    if do_plot: sim.plot()

    assert res.cum_infections[-1] > res.cum_infections[0] # There were infections
    return sim


@sc.timer()
def test_single_uids(do_plot=do_plot):
    """ Test a single MixingPool by UIDS, pre-initialization configuration """
    test_name = sys._getframe().f_code.co_name
    sc.heading(f'Testing {test_name}...')

    # Note: in this arrangement, src transmits to dst, but can never go the other way, so the epidemic dies out once everyone in src recovers
    k = pars.n_agents // 2
    mp = ss.MixingPool(
        src = ss.uids(np.arange(k)),
        dst = ss.uids(np.arange(k, pars.n_agents))
    )
    sir = ss.SIR()
    sim = ss.Sim(pars, diseases=sir, networks=mp, label=test_name)
    sim.run()

    if do_plot: sim.plot()

    assert sim.results.sir['cum_infections'][-1] > sim.results.sir['cum_infections'][0], 'There were no new infections in the simulation'
    return sim


@sc.timer()
def test_single_ncd():
    """ Test a single MixingPool with a ncd """
    test_name = sys._getframe().f_code.co_name
    sc.heading(f'Testing {test_name}...')
    mp_pars = dict(
        src = ss.AgeGroup(0, 15),
        dst = ss.AgeGroup(15, None),
        beta = 1,
        n_contacts = ss.poisson(lam=5),
        diseases = 'ncd'
    )
    mp = ss.MixingPool(**mp_pars)

    ncd = ss.NCD()
    sim = ss.Sim(pars, diseases=ncd, networks=mp, label=test_name)
    with pytest.raises(Exception):
        sim.run()
    return sim


@sc.timer()
def test_single_missing_disease():
    """ Test a single MixingPool with a missing disease """
    test_name = sys._getframe().f_code.co_name
    sc.heading(f'Testing {test_name}...')
    mp_pars = dict(
        src = ss.AgeGroup(0, 15),
        dst = ss.AgeGroup(15, None),
        beta = 1.0,
        n_contacts = ss.poisson(lam=5),
        diseases = 'hiv'
    )
    mp = ss.MixingPool(**mp_pars)

    sir = ss.SIR()
    sim = ss.Sim(pars, diseases=sir, networks=mp, label=test_name)
    with pytest.raises(Exception):
        sim.run()

    return sim


@sc.timer()
def test_single_age(do_plot=do_plot):
    """ Test a single MixingPool by age """
    # Incidence must decline because 0-15 --> 15+ transmission only
    test_name = sys._getframe().f_code.co_name
    sc.heading(f'Testing {test_name}...')
    mp_pars = dict(
        src = ss.AgeGroup(0, 15),
        dst = ss.AgeGroup(15, None),
        beta = 1.0,
        n_contacts = ss.poisson(lam=5),
    )
    mp = ss.MixingPool(**mp_pars)

    sir = ss.SIR()
    sim = ss.Sim(pars, diseases=sir, networks=mp, label=test_name)
    sim.run()

    if do_plot: sim.plot()

    assert(sim.results.sir['cum_infections'][-1] > sim.results.sir['cum_infections'][0]) # There were infections
    return sim


@sc.timer()
def test_single_sex(do_plot=do_plot):
    """ Test a single MixingPool by sex """
    # Incidence must decline because M --> F transmission only
    test_name = sys._getframe().f_code.co_name
    sc.heading(f'Testing {test_name}...')
    mp_pars = dict(
        src = lambda sim: sim.people.female, # female to male (only) transmission
        dst = lambda sim: sim.people.male,
        beta = 1.0,
        n_contacts = ss.poisson(lam=4),
    )
    mp = ss.MixingPool(**mp_pars)

    sir = ss.SIR(init_prev=ss.bernoulli(p=lambda self, sim, uids: 0.05*sim.people.female)) # Seed 5% of the female population
    sim = ss.Sim(pars, diseases=sir, networks=mp, label=test_name)
    sim.run()

    if do_plot: sim.plot()

    assert(sim.results.sir['cum_infections'][-1] > sim.results.sir['cum_infections'][0]) # There were infections
    assert(sim.people.male[sim.diseases.sir.ti_infected>0].all()) # All new infections should be in men
    return sim


@sc.timer()
def test_multi_defaults(do_plot=do_plot):
    """ Test MixingPools using defaults """
    test_name = sys._getframe().f_code.co_name
    sc.heading(f'Testing {test_name}...')
    mps = ss.MixingPools(src={'all':None}, dst={'all':None}, n_contacts=[[1]])
    sir = ss.SIR()
    sim = ss.Sim(pars, diseases=sir, networks=mps, label=test_name)
    sim.run()

    if do_plot: sim.plot()

    assert(sim.results.sir['cum_infections'][-1] > sim.results.sir['cum_infections'][0]) # There were infections
    return sim


@sc.timer()
def test_multi(do_plot=do_plot):
    """ Test MixingPools """
    test_name = sys._getframe().f_code.co_name
    sc.heading(f'Testing {test_name}...')

    groups = {
        'Female': lambda sim: sim.people.female,
        'Male': lambda sim: sim.people.male,
    }

    mps_pars = dict(
        n_contacts = np.array([[1.4, 0.5], [1.2, 0.7]]),
        beta = 1.0,
        src = groups,
        dst = groups,
    )
    mps = ss.MixingPools(**mps_pars)

    sir = ss.SIR()
    sim = ss.Sim(pars, diseases=sir, networks=mps, label=test_name)
    sim.run()

    if do_plot: sim.plot()

    assert(sim.results.sir['cum_infections'][-1] > sim.results.sir['cum_infections'][0]) # There were infections
    return sim


if __name__ == '__main__':
    do_plot = True
    sc.options(interactive=do_plot)
    T = sc.timer()

    sim0 = test_single_defaults(do_plot)
    sim1 = test_single_uids(do_plot)
    sim2 = test_single_ncd()
    sim3 = test_single_missing_disease()
    sim4 = test_single_age(do_plot)
    sim5 = test_single_sex(do_plot)

    sim6 = test_multi_defaults(do_plot)
    sim7 = test_multi(do_plot)

    T.toc()

    if do_plot:
        plt.show()
